#! /usr/bin/env python
"""Service parsing metadata from PDFs/TEIs (created by GROBID)."""
from __future__ import annotations

import re
from pathlib import Path
from typing import Optional
from xml import etree
from xml.etree.ElementTree import Element  # nosec

import defusedxml
import requests
from defusedxml.ElementTree import fromstring
from lxml.etree import XMLSyntaxError  # nosec

import colrev.env.grobid_service
import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.record

# xpath alternative:
# tree.xpath("//tei:sourceDesc/tei:biblStruct/tei:monogr/tei:idno",
# namespaces={"tei": "http://www.tei-c.org/ns/1.0"})
# abstract_node =tree.xpath("//tei:profileDesc/tei:abstract",
# namespaces={"tei": "http://www.tei-c.org/ns/1.0"})
# etree.tostring(abstract_node[0]).decode("utf-8")


# defuse std xml lib
defusedxml.defuse_stdlib()


class TEIParser:
    """Environment service for TEI parsing"""

    ns = {
        "tei": "{http://www.tei-c.org/ns/1.0}",
        "w3": "{http://www.w3.org/XML/1998/namespace}",
    }
    nsmap = {
        "tei": "http://www.tei-c.org/ns/1.0",
        "w3": "http://www.w3.org/XML/1998/namespace",
    }

    def __init__(
        self,
        *,
        environment_manager: colrev.env.environment_manager.EnvironmentManager,
        pdf_path: Optional[Path] = None,
        tei_path: Optional[Path] = None,
    ):
        """Creates a TEI file
        modes of operation:
        - pdf_path: create TEI and temporarily store in self.data
        - pfd_path and tei_path: create TEI and save in tei_path
        - tei_path: read TEI from file
        """

        self.environment_manager = environment_manager
        # pylint: disable=consider-using-with
        assert pdf_path is not None or tei_path is not None
        if pdf_path is not None:
            if pdf_path.is_symlink():
                pdf_path = pdf_path.resolve()
        self.pdf_path = pdf_path
        self.tei_path = tei_path
        if pdf_path is not None and not pdf_path.is_file():
            raise FileNotFoundError

        load_from_tei = False
        if tei_path is not None:
            if tei_path.is_file():
                load_from_tei = True

        if pdf_path is not None and not load_from_tei:
            self.__create_tei()

        elif tei_path is not None:
            self.root = self.__read_from_tei()  # type: ignore

    def __read_from_tei(self):  # type: ignore
        """Read a TEI from file"""
        with open(self.tei_path, "rb") as data:
            xslt_content = data.read()

        if b"[BAD_INPUT_DATA]" in xslt_content[:100]:
            raise colrev_exceptions.TEIException()

        return etree.ElementTree.XML(xslt_content)

    def __create_tei(self) -> None:
        """Create the TEI (based on GROBID)"""
        grobid_service = colrev.env.grobid_service.GrobidService(
            environment_manager=self.environment_manager
        )
        grobid_service.start()
        # Note: we have more control and transparency over the consolidation
        # if we do it in the colrev process
        options = {"consolidateHeader": "0", "consolidateCitations": "0"}

        # Note: Grobid offers direct export of Bibtex:
        # r = requests.post(
        #     GROBID_SERVICE.GROBID_URL() + "/api/processHeaderDocument",
        #     headers={"Accept": "application/x-bibtex"},
        # But parsing the metadata from the tei gives us more control of the details

        try:
            # pylint: disable=consider-using-with
            ret = requests.post(
                grobid_service.GROBID_URL + "/api/processFulltextDocument",
                files={"input": open(str(self.pdf_path), "rb")},
                data=options,
                timeout=180,
            )

            # Possible extension: get header only (should be more efficient)
            # r = requests.post(
            #     GrobidService.GROBID_URL + "/api/processHeaderDocument",
            #     files=dict(input=open(filepath, "rb")),
            #     data=header_data,
            # )

            if ret.status_code != 200:
                raise colrev_exceptions.TEIException()

            if b"[TIMEOUT]" in ret.content:
                raise colrev_exceptions.TEITimeoutException()

            self.root = fromstring(ret.content)

            if self.tei_path is not None:
                self.tei_path.parent.mkdir(exist_ok=True, parents=True)
                with open(self.tei_path, "wb") as file:
                    file.write(ret.content)

                # Note : reopen/write to prevent format changes in the enhancement
                with open(self.tei_path, "rb") as file:
                    xml_fstring = file.read()
                self.root = fromstring(xml_fstring)

                tree = etree.ElementTree.ElementTree(self.root)
                tree.write(str(self.tei_path), encoding="utf-8")
        except requests.exceptions.ConnectionError as exc:
            print(exc)
            print(str(self.pdf_path))

    def get_tei_str(self) -> str:
        """Get the TEI string"""
        try:
            etree.ElementTree.register_namespace("tei", "http://www.tei-c.org/ns/1.0")
            return etree.ElementTree.tostring(self.root).decode("utf-8")
        except XMLSyntaxError as exc:
            raise colrev_exceptions.TEIException from exc

    def get_grobid_version(self) -> str:
        """Get the GROBID version used for TEI creation"""
        grobid_version = "NA"
        encoding_description = self.root.find(".//" + self.ns["tei"] + "encodingDesc")
        if encoding_description is not None:
            app_info_node = encoding_description.find(
                ".//" + self.ns["tei"] + "appInfo"
            )
            if app_info_node is not None:
                application_node = encoding_description.find(
                    ".//" + self.ns["tei"] + "application"
                )
                if application_node is not None:
                    if application_node.get("version") is not None:
                        grobid_version = application_node.get("version")
        return grobid_version

    def __get_paper_title(self) -> str:
        title_text = "NA"
        file_description = self.root.find(".//" + self.ns["tei"] + "fileDesc")
        if file_description is not None:
            title_stmt_node = file_description.find(
                ".//" + self.ns["tei"] + "titleStmt"
            )
            if title_stmt_node is not None:
                title_node = title_stmt_node.find(".//" + self.ns["tei"] + "title")
                if title_node is not None:
                    title_text = (
                        title_node.text if title_node.text is not None else "NA"
                    )
                    title_text = (
                        title_text.replace("(Completed paper)", "")
                        .replace("(Completed-paper)", "")
                        .replace("(Research-in-Progress)", "")
                        .replace("Completed Research Paper", "")
                    )
        return title_text

    def __get_paper_journal(self) -> str:
        # pylint: disable=too-many-nested-blocks
        journal_name = "NA"
        file_description = self.root.find(".//" + self.ns["tei"] + "sourceDesc")
        if file_description is not None:
            if file_description.find(".//" + self.ns["tei"] + "monogr") is not None:
                journal_node = file_description.find(".//" + self.ns["tei"] + "monogr")
                if journal_node is not None:
                    jtitle_node = journal_node.find(".//" + self.ns["tei"] + "title")
                    if jtitle_node is not None:
                        journal_name = (
                            jtitle_node.text if jtitle_node.text is not None else "NA"
                        )
                        if journal_name != "NA":
                            words = journal_name.split()
                            if sum(word.isupper() for word in words) / len(words) > 0.8:
                                words = [word.capitalize() for word in words]
                                journal_name = " ".join(words)
        return journal_name

    def __get_paper_journal_volume(self) -> str:
        volume = "NA"
        file_description = self.root.find(".//" + self.ns["tei"] + "sourceDesc")
        if file_description is not None:
            if file_description.find(".//" + self.ns["tei"] + "monogr") is not None:
                journal_node = file_description.find(".//" + self.ns["tei"] + "monogr")
                if journal_node is not None:
                    imprint_node = journal_node.find(".//" + self.ns["tei"] + "imprint")
                    if imprint_node is not None:
                        vnode = imprint_node.find(
                            ".//" + self.ns["tei"] + "biblScope[@unit='volume']"
                        )
                        if vnode is not None:
                            volume = vnode.text if vnode.text is not None else "NA"
        return volume

    def __get_paper_journal_issue(self) -> str:
        issue = "NA"
        file_description = self.root.find(".//" + self.ns["tei"] + "sourceDesc")
        if file_description is not None:
            if file_description.find(".//" + self.ns["tei"] + "monogr") is not None:
                journal_node = file_description.find(".//" + self.ns["tei"] + "monogr")
                if journal_node is not None:
                    imprint_node = journal_node.find(".//" + self.ns["tei"] + "imprint")
                    if imprint_node is not None:
                        issue_node = imprint_node.find(
                            ".//" + self.ns["tei"] + "biblScope[@unit='issue']"
                        )
                        if issue_node is not None:
                            issue = (
                                issue_node.text if issue_node.text is not None else "NA"
                            )
        return issue

    def __get_paper_journal_pages(self) -> str:
        pages = "NA"
        file_description = self.root.find(".//" + self.ns["tei"] + "sourceDesc")
        if file_description is not None:
            journal_node = file_description.find(".//" + self.ns["tei"] + "monogr")
            if journal_node is not None:
                imprint_node = journal_node.find(".//" + self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    page_node = imprint_node.find(
                        ".//" + self.ns["tei"] + "biblScope[@unit='page']"
                    )
                    if page_node is not None:
                        if (
                            page_node.get("from") is not None
                            and page_node.get("to") is not None
                        ):
                            pages = (
                                page_node.get("from", "")
                                + "--"
                                + page_node.get("to", "")
                            )
        return pages

    def __get_paper_year(self) -> str:
        year = "NA"
        file_description = self.root.find(".//" + self.ns["tei"] + "sourceDesc")
        if file_description is not None:
            if file_description.find(".//" + self.ns["tei"] + "monogr") is not None:
                journal_node = file_description.find(".//" + self.ns["tei"] + "monogr")
                if journal_node is not None:
                    imprint_node = journal_node.find(".//" + self.ns["tei"] + "imprint")
                    if imprint_node is not None:
                        date_node = imprint_node.find(".//" + self.ns["tei"] + "date")
                        if date_node is not None:
                            year = (
                                date_node.get("when", "")
                                if date_node.get("when") is not None
                                else "NA"
                            )
                            year = re.sub(r".*([1-2][0-9]{3}).*", r"\1", year)
        return year

    def __parse_author_dict(self, *, author_pers_node: Element) -> dict:
        author_dict = {}
        surname_node = author_pers_node.find(self.ns["tei"] + "surname")
        if surname_node is not None:
            surname = surname_node.text if surname_node.text is not None else ""
            author_dict["surname"] = surname
        else:
            author_dict["surname"] = ""

        forename_node = author_pers_node.find(
            self.ns["tei"] + 'forename[@type="first"]'
        )
        if forename_node is not None:
            forename = forename_node.text if forename_node.text is not None else ""
            if 1 == len(forename):
                forename = forename + "."
            author_dict["forename"] = forename
        else:
            author_dict["forename"] = ""

        middlename_node = author_pers_node.find(
            self.ns["tei"] + 'forename[@type="middle"]'
        )
        if middlename_node is not None:
            middlename = (
                " " + middlename_node.text if middlename_node.text is not None else ""
            )
            if 1 == len(middlename):
                middlename = middlename + "."
            author_dict["middlename"] = middlename
        else:
            middlename = ""

        return author_dict

    def __get_author_name_from_node(self, *, author_node: Element) -> str:
        authorname = ""

        author_pers_node = author_node.find(self.ns["tei"] + "persName")
        if author_pers_node is None:
            return authorname

        author_dict = self.__parse_author_dict(author_pers_node=author_pers_node)

        authorname = author_dict["surname"] + ", " + author_dict["forename"]
        if "middlename" in author_dict:
            authorname += author_dict["middlename"]

        authorname = (
            authorname.replace("\n", " ")
            .replace("\r", "")
            .replace("•", "")
            .replace("+", "")
            .replace("Dipl.", "")
            .replace("Prof.", "")
            .replace("Dr.", "")
            .replace("&apos", "'")
            .replace("❚", "")
            .replace("~", "")
            .replace("®", "")
            .replace("|", "")
        )

        authorname = re.sub("^Paper, Short; ", "", authorname)
        return authorname

    def __get_paper_authors(self) -> str:
        author_string = "NA"
        file_description = self.root.find(".//" + self.ns["tei"] + "sourceDesc")
        author_list = []

        if file_description is not None:
            if file_description.find(".//" + self.ns["tei"] + "analytic") is not None:
                analytic_node = file_description.find(
                    ".//" + self.ns["tei"] + "analytic"
                )
                if analytic_node is not None:
                    for author_node in analytic_node.iterfind(
                        self.ns["tei"] + "author"
                    ):
                        authorname = self.__get_author_name_from_node(
                            author_node=author_node
                        )
                        if authorname in ["Paper, Short"]:
                            continue
                        if authorname not in [", ", ""]:
                            author_list.append(authorname)

                    author_string = " and ".join(author_list)

                    if author_string is None:
                        author_string = "NA"
                    if "" == author_string.replace(" ", "").replace(",", "").replace(
                        ";", ""
                    ):
                        author_string = "NA"
        return author_string

    def __get_paper_doi(self) -> str:
        doi = "NA"
        file_description = self.root.find(".//" + self.ns["tei"] + "sourceDesc")
        if file_description is not None:
            bibl_struct = file_description.find(".//" + self.ns["tei"] + "biblStruct")
            if bibl_struct is not None:
                dois = bibl_struct.findall(".//" + self.ns["tei"] + "idno[@type='DOI']")
                for res in dois:
                    if res.text is not None:
                        doi = res.text
        return doi

    def get_abstract(self) -> str:
        """Get the abstract"""

        html_tag_regex = re.compile("<.*?>")

        def cleanhtml(raw_html: str) -> str:
            cleantext = re.sub(html_tag_regex, "", raw_html)
            return cleantext

        abstract_text = "NA"
        profile_description = self.root.find(".//" + self.ns["tei"] + "profileDesc")
        if profile_description is not None:
            abstract_node = profile_description.find(
                ".//" + self.ns["tei"] + "abstract"
            )
            html_str = etree.ElementTree.tostring(abstract_node).decode("utf-8")
            abstract_text = cleanhtml(html_str)
        abstract_text = abstract_text.lstrip().rstrip()
        return abstract_text

    def get_metadata(self) -> dict:
        """Get the metadata of the PDF (title, author, ...) as a dict"""

        record = {
            "ENTRYTYPE": "article",
            "title": self.__get_paper_title(),
            "author": self.__get_paper_authors(),
            "journal": self.__get_paper_journal(),
            "year": self.__get_paper_year(),
            "volume": self.__get_paper_journal_volume(),
            "number": self.__get_paper_journal_issue(),
            "pages": self.__get_paper_journal_pages(),
            "doi": self.__get_paper_doi(),
        }

        for key, value in record.items():
            if key != "file":
                record[key] = value.replace("}", "").replace("{", "").rstrip("\\")
            else:
                print(f"problem in filename: {key}")

        return record

    def get_paper_keywords(self) -> list:
        """Get hte keywords"""
        keywords = []
        for keyword_list in self.root.iter(self.ns["tei"] + "keywords"):
            for keyword in keyword_list.iter(self.ns["tei"] + "term"):
                keywords.append(keyword.text)
        return keywords

    def get_author_details(self) -> list:
        """Get the author details"""
        author_details = []

        file_description = self.root.find(".//" + self.ns["tei"] + "sourceDesc")

        if file_description is not None:
            if file_description.find(".//" + self.ns["tei"] + "analytic") is not None:
                analytic_node = file_description.find(
                    ".//" + self.ns["tei"] + "analytic"
                )
                if analytic_node is not None:
                    for author_node in analytic_node.iterfind(
                        self.ns["tei"] + "author"
                    ):
                        author_pers_node = author_node.find(self.ns["tei"] + "persName")
                        if author_pers_node is None:
                            continue

                        author_dict = self.__parse_author_dict(
                            author_pers_node=author_pers_node
                        )

                        email_node = author_node.find(self.ns["tei"] + "email")
                        if email_node is not None:
                            author_dict["emai"] = email_node.text

                        orcid_node = author_node.find(
                            self.ns["tei"] + 'idno[@type="ORCID"]'
                        )
                        if orcid_node is not None:
                            orcid = orcid_node.text
                            author_dict["ORCID"] = orcid

                        author_details.append(author_dict)

        return author_details

    # (individual) bibliography-reference elements  ----------------------------

    def __get_reference_bibliography_tei_id(self, *, reference: Element) -> str:
        return reference.attrib[self.ns["w3"] + "id"]

    def __get_reference_author_string(self, *, reference: Element) -> str:
        author_list = []
        if reference.find(self.ns["tei"] + "analytic") is not None:
            authors_node = reference.find(self.ns["tei"] + "analytic")
        elif reference.find(self.ns["tei"] + "monogr") is not None:
            authors_node = reference.find(self.ns["tei"] + "monogr")

        if authors_node is not None:
            for author_node in authors_node.iterfind(self.ns["tei"] + "author"):
                authorname = self.__get_author_name_from_node(author_node=author_node)

                if authorname not in [", ", ""]:
                    author_list.append(authorname)

        author_string = " and ".join(author_list)

        author_string = (
            author_string.replace("\n", " ")
            .replace("\r", "")
            .replace("•", "")
            .replace("+", "")
            .replace("Dipl.", "")
            .replace("Prof.", "")
            .replace("Dr.", "")
            .replace("&apos", "'")
            .replace("❚", "")
            .replace("~", "")
            .replace("®", "")
            .replace("|", "")
        )

        if author_string is None:
            author_string = "NA"
        if author_string.replace(" ", "").replace(",", "").replace(";", "") == "":
            author_string = "NA"
        return author_string

    def __get_reference_title_string(self, *, reference: Element) -> str:
        title_string = ""
        if reference.find(self.ns["tei"] + "analytic") is not None:
            analytic_node = reference.find(self.ns["tei"] + "analytic")
            if analytic_node is not None:
                title = analytic_node.find(self.ns["tei"] + "title")
        elif reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                title = monogr_node.find(self.ns["tei"] + "title")

        if title is not None:
            if title.text is not None:
                title_string = title.text

        return title_string

    def __get_reference_year_string(self, *, reference: Element) -> str:
        year_string = ""
        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                imprint_node = monogr_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    year = imprint_node.find(self.ns["tei"] + "date")
        elif reference.find(self.ns["tei"] + "analytic") is not None:
            analytic_node = reference.find(self.ns["tei"] + "analytic")
            if analytic_node is not None:
                imprint_node = analytic_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    year = imprint_node.find(self.ns["tei"] + "date")

        if year is not None:
            for name, value in sorted(year.items()):
                year_string = value if (name == "when") else "NA"
        else:
            year_string = "NA"
        return year_string

    def __get_reference_page_string(self, *, reference: Element) -> str:
        page_string = ""

        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                imprint_node = monogr_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    page_list = imprint_node.findall(
                        self.ns["tei"] + "biblScope[@unit='page']"
                    )
        elif reference.find(self.ns["tei"] + "analytic") is not None:
            analytic_node = reference.find(self.ns["tei"] + "analytic")
            if analytic_node is not None:
                imprint_node = analytic_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    page_list = imprint_node.findall(
                        self.ns["tei"] + "biblScope[@unit='page']"
                    )

        for page in page_list:
            if page is not None:
                for name, value in sorted(page.items()):
                    if name == "from":
                        page_string += value
                    if name == "to":
                        page_string += "--" + value
            else:
                page_string = "NA"

        return page_string

    def __get_reference_number_string(self, *, reference: Element) -> str:
        number_string = ""

        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                imprint_node = monogr_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    number_list = imprint_node.findall(
                        self.ns["tei"] + "biblScope[@unit='issue']"
                    )
        elif reference.find(self.ns["tei"] + "analytic") is not None:
            analytic_node = reference.find(self.ns["tei"] + "analytic")
            if analytic_node is not None:
                imprint_node = analytic_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    number_list = imprint_node.findall(
                        self.ns["tei"] + "biblScope[@unit='issue']"
                    )

        for number in number_list:
            if number.text is not None:
                number_string = number.text

        return number_string

    def __get_reference_volume_string(self, *, reference: Element) -> str:
        volume_string = ""

        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                imprint_node = monogr_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    volume_list = imprint_node.findall(
                        self.ns["tei"] + "biblScope[@unit='volume']"
                    )
        elif reference.find(self.ns["tei"] + "analytic") is not None:
            analytic_node = reference.find(self.ns["tei"] + "analytic")
            if analytic_node is not None:
                imprint_node = analytic_node.find(self.ns["tei"] + "imprint")
                if imprint_node is not None:
                    volume_list = imprint_node.findall(
                        self.ns["tei"] + "biblScope[@unit='volume']"
                    )

        for volume in volume_list:
            if volume.text is not None:
                volume_string = volume.text

        return volume_string

    def __get_reference_journal_string(self, *, reference: Element) -> str:
        journal_title = ""
        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                monogr_title = monogr_node.find(self.ns["tei"] + "title")
                if monogr_title is not None:
                    if monogr_title.text is not None:
                        journal_title = monogr_title.text

        return journal_title

    def __get_entrytype(self, *, reference: Element) -> str:
        entrytype = "misc"
        if reference.find(self.ns["tei"] + "monogr") is not None:
            monogr_node = reference.find(self.ns["tei"] + "monogr")
            if monogr_node is not None:
                title_node = monogr_node.find(self.ns["tei"] + "title")
                if title_node is not None:
                    if title_node.get("level", "NA") != "j":
                        entrytype = "book"
                    else:
                        entrytype = "article"
        return entrytype

    def __get_tei_id_count(self, *, tei_id: str) -> int:
        count = 0

        for reference in self.root.iter(self.ns["tei"] + "ref"):
            if "target" in reference.keys():
                if reference.get("target") == f"#{tei_id}":
                    count += 1

        return count

    def get_bibliography(self, *, min_intext_citations: int = 0) -> list:
        """Get the bibliography (references section) as a list of record dicts"""
        # Note : could also allow top-10 % of most frequent in-text citations

        bibliographies = self.root.iter(self.ns["tei"] + "listBibl")
        tei_bib_db = []
        for bibliography in bibliographies:
            for reference in bibliography:
                try:
                    entrytype = self.__get_entrytype(reference=reference)
                    tei_id = self.__get_reference_bibliography_tei_id(
                        reference=reference
                    )

                    if min_intext_citations > 0:
                        if (
                            self.__get_tei_id_count(tei_id=tei_id)
                            < min_intext_citations
                        ):
                            continue

                    if entrytype == "article":
                        ref_rec = {
                            "ID": tei_id,
                            "ENTRYTYPE": entrytype,
                            "tei_id": tei_id,
                            "author": self.__get_reference_author_string(
                                reference=reference
                            ),
                            "title": self.__get_reference_title_string(
                                reference=reference
                            ),
                            "year": self.__get_reference_year_string(
                                reference=reference
                            ),
                            "journal": self.__get_reference_journal_string(
                                reference=reference
                            ),
                            "volume": self.__get_reference_volume_string(
                                reference=reference
                            ),
                            "number": self.__get_reference_number_string(
                                reference=reference
                            ),
                            "pages": self.__get_reference_page_string(
                                reference=reference
                            ),
                        }
                    elif entrytype == "book":
                        ref_rec = {
                            "ID": tei_id,
                            "ENTRYTYPE": entrytype,
                            "tei_id": tei_id,
                            "author": self.__get_reference_author_string(
                                reference=reference
                            ),
                            "title": self.__get_reference_title_string(
                                reference=reference
                            ),
                            "year": self.__get_reference_year_string(
                                reference=reference
                            ),
                        }
                    elif entrytype == "misc":
                        ref_rec = {
                            "ID": tei_id,
                            "ENTRYTYPE": entrytype,
                            "tei_id": tei_id,
                            "author": self.__get_reference_author_string(
                                reference=reference
                            ),
                            "title": self.__get_reference_title_string(
                                reference=reference
                            ),
                        }
                except etree.ElementTree.ParseError:
                    continue

                ref_rec = {k: v for k, v in ref_rec.items() if v is not None}
                # print(ref_rec)
                tei_bib_db.append(ref_rec)

        return tei_bib_db

    def get_citations_per_section(self) -> dict:
        """Get a dict of section-names and list-of-citations"""
        section_citations = {}
        parent_map = {c: p for p in self.root.iter() for c in p}
        sections = self.root.iter(f'{self.ns["tei"]}head')
        for section in sections:
            section_name = section.text
            if section_name is None:
                continue
            parent = parent_map[section]
            citation_nodes = parent.findall(f'.//{self.ns["tei"]}ref')
            citations = [
                x.get("target", "NA").replace("#", "")
                for x in citation_nodes
                if x.get("type", "NA") == "bibr"
            ]
            citations = list(filter(lambda a: a != "NA", citations))
            if len(citations) > 0:
                section_citations[section_name.lower()] = citations
        return section_citations

    def mark_references(self, *, records: dict):  # type: ignore
        """Mark references with the additional record ID"""

        tei_records = self.get_bibliography()
        for record_dict in tei_records:
            if "title" not in record_dict:
                continue

            max_sim = 0.9
            max_sim_record = {}
            for local_record_dict in records.values():
                if local_record_dict["colrev_status"] not in [
                    colrev.record.RecordState.rev_included,
                    colrev.record.RecordState.rev_synthesized,
                ]:
                    continue
                rec_sim = colrev.record.Record.get_record_similarity(
                    record_a=colrev.record.Record(data=record_dict),
                    record_b=colrev.record.Record(data=local_record_dict),
                )
                if rec_sim > max_sim:
                    max_sim_record = local_record_dict
                    max_sim = rec_sim
            if len(max_sim_record) == 0:
                continue

            # Record found: mark in tei
            bibliography = self.root.find(f".//{self.ns['tei']}listBibl")
            # mark reference in bibliography
            for ref in bibliography:
                if ref.get(f'{self.ns["w3"]}id') == record_dict["tei_id"]:
                    ref.set("ID", max_sim_record["ID"])
            # mark reference in in-text citations
            for reference in self.root.iter(f'{self.ns["tei"]}ref'):
                if "target" in reference.keys():
                    if reference.get("target") == f"#{record_dict['tei_id']}":
                        reference.set("ID", max_sim_record["ID"])

            # if settings file available: dedupe_io match agains records

        if self.tei_path:
            tree = etree.ElementTree.ElementTree(self.root)
            tree.write(str(self.tei_path))

        return self.root



#!/usr/bin/env python3
"""Types and model of CoLRev operations."""
from __future__ import annotations

import typing
from enum import auto
from enum import Enum
from typing import Any
from typing import Callable
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar

import docker
import git
from docker.errors import DockerException

import colrev.exceptions as colrev_exceptions
import colrev.record

if TYPE_CHECKING:
    import colrev.review_manager

F = TypeVar("F", bound=Callable[..., Any])


class OperationsType(Enum):
    """Operation types correspond to the main state transitions (see RecordStateModel)"""

    # pylint: disable=invalid-name

    search = auto()
    load = auto()
    prep = auto()
    prep_man = auto()
    dedupe = auto()
    prescreen = auto()
    pdf_get = auto()
    pdf_get_man = auto()
    pdf_prep = auto()
    pdf_prep_man = auto()
    screen = auto()
    data = auto()

    format = auto()
    check = auto()

    def __str__(self) -> str:
        return f"{self.name}"


class Operation:
    """Operations correspond to the work steps in a CoLRev project"""

    # pylint: disable=too-few-public-methods

    force_mode: bool
    type: OperationsType

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        operations_type: OperationsType,
        notify_state_transition_operation: bool = True,
    ) -> None:
        self.review_manager = review_manager
        self.force_mode = self.review_manager.force_mode

        self.type = operations_type

        self.notify_state_transition_operation = notify_state_transition_operation
        if notify_state_transition_operation:
            self.review_manager.notify(operation=self)
        else:
            self.review_manager.notify(operation=self, state_transition=False)

        self.cpus = 4

        self.docker_images_to_stop: typing.List[str] = []

        # Note: the following call seems to block the flow (if debug is enabled)
        # self.review_manager.logger.debug(f"Created {self.type} operation")

        # Note: we call review_manager.notify() in the subclasses
        # to make sure that the review_manager calls the right check_preconditions()

    # pylint: disable=too-many-nested-blocks
    @classmethod
    def decorate(cls) -> Callable:
        """Decorator for operations"""

        def decorator_func(func: F) -> Callable:
            def wrapper_func(self, *args, **kwargs) -> Any:  # type: ignore
                # Invoke the wrapped function
                retval = func(self, *args, **kwargs)
                # Conclude the operation
                self.conclude()
                if self.review_manager.in_ci_environment():
                    print("\n\n")
                return retval

            return wrapper_func

        return decorator_func

    def __check_record_state_model_precondition(self) -> None:
        colrev.record.RecordStateModel.check_operation_precondition(operation=self)

    def __require_clean_repo_general(
        self,
        *,
        git_repo: Optional[git.Repo] = None,
        ignore_pattern: Optional[list] = None,
    ) -> bool:
        if git_repo is None:
            git_repo = git.Repo(self.review_manager.path)

        # Note : not considering untracked files.

        if len(git_repo.index.diff("HEAD")) == 0:
            unstaged_changes = [item.a_path for item in git_repo.index.diff(None)]
            if self.review_manager.dataset.RECORDS_FILE_RELATIVE in unstaged_changes:
                git_repo.index.add([self.review_manager.dataset.RECORDS_FILE_RELATIVE])

        # Principle: working tree always has to be clean
        # because processing functions may change content
        if git_repo.is_dirty(index=False):
            changed_files = [item.a_path for item in git_repo.index.diff(None)]
            raise colrev_exceptions.UnstagedGitChangesError(changed_files)

        if git_repo.is_dirty():
            if ignore_pattern is None:
                changed_files = [item.a_path for item in git_repo.index.diff(None)] + [
                    x.a_path
                    for x in git_repo.head.commit.diff()
                    if x.a_path not in [str(self.review_manager.STATUS_RELATIVE)]
                ]
                if len(changed_files) > 0:
                    raise colrev_exceptions.CleanRepoRequiredError(changed_files, "")
            else:
                changed_files = [
                    item.a_path
                    for item in git_repo.index.diff(None)
                    if not any(str(ip) in item.a_path for ip in ignore_pattern)
                ] + [
                    x.a_path
                    for x in git_repo.head.commit.diff()
                    if not any(str(ip) in x.a_path for ip in ignore_pattern)
                ]
                if str(self.review_manager.STATUS_RELATIVE) in changed_files:
                    changed_files.remove(str(self.review_manager.STATUS_RELATIVE))
                if changed_files:
                    raise colrev_exceptions.CleanRepoRequiredError(
                        changed_files, ",".join([str(x) for x in ignore_pattern])
                    )
        return True

    def check_precondition(self) -> None:
        """Check the operation precondition"""

        if self.force_mode:
            return

        if OperationsType.load == self.type:
            self.__require_clean_repo_general(
                ignore_pattern=[
                    self.review_manager.SEARCHDIR_RELATIVE,
                    self.review_manager.SETTINGS_RELATIVE,
                ]
            )
            self.__check_record_state_model_precondition()

        elif OperationsType.prep == self.type:
            if self.notify_state_transition_operation:
                self.__require_clean_repo_general()
                self.__check_record_state_model_precondition()

        elif OperationsType.prep_man == self.type:
            self.__require_clean_repo_general(
                ignore_pattern=[self.review_manager.dataset.RECORDS_FILE_RELATIVE]
            )
            self.__check_record_state_model_precondition()

        elif OperationsType.dedupe == self.type:
            self.__require_clean_repo_general()
            self.__check_record_state_model_precondition()

        elif OperationsType.prescreen == self.type:
            self.__require_clean_repo_general()
            self.__check_record_state_model_precondition()

        elif OperationsType.pdf_get == self.type:
            self.__require_clean_repo_general(
                ignore_pattern=[self.review_manager.PDF_DIR_RELATIVE]
            )
            self.__check_record_state_model_precondition()

        elif OperationsType.pdf_get_man == self.type:
            self.__require_clean_repo_general(
                ignore_pattern=[self.review_manager.PDF_DIR_RELATIVE]
            )
            self.__check_record_state_model_precondition()

        elif OperationsType.pdf_prep == self.type:
            self.__require_clean_repo_general()
            self.__check_record_state_model_precondition()

        elif OperationsType.screen == self.type:
            self.__require_clean_repo_general()
            self.__check_record_state_model_precondition()

        elif OperationsType.data == self.type:
            # __require_clean_repo_general(
            #     ignore_pattern=[
            #         # data.csv, paper.md etc.?,
            #     ]
            # )
            self.__check_record_state_model_precondition()

        # ie., implicit pass for format, explore, check, pdf_prep_man

    def conclude(self) -> None:
        """Conclude the operation (stop Docker containers)"""
        try:
            client = docker.from_env()
            for container in client.containers.list():
                if any(x in container.image.tags for x in self.docker_images_to_stop):
                    container.stop()
        except DockerException:
            pass


class FormatOperation(Operation):
    """A dummy operation that is expected to introduce formatting changes only"""

    # pylint: disable=too-few-public-methods

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager, operations_type=OperationsType.format
        )
        if notify:
            self.review_manager.notify(operation=self)


class CheckOperation(Operation):
    """A dummy operation that is not expected to introduce changes"""

    # pylint: disable=too-few-public-methods

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=OperationsType.check,
            notify_state_transition_operation=False,
        )



#! /usr/bin/env python
"""Checker for fields that are inconsistent with the ENTRYTYPE."""
from __future__ import annotations

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class InconsistentWithEntrytypeChecker:
    """The InconsistentWithEntrytypeChecker"""

    record_field_inconsistencies: dict[str, list[str]] = {
        "article": ["booktitle", "isbn"],
        "inproceedings": ["issue", "number", "journal", "isbn"],
        "incollection": [],
        "inbook": ["journal"],
        "book": ["issue", "number", "journal"],
        "phdthesis": ["volume", "issue", "number", "journal", "booktitle", "isbn"],
        "masterthesis": ["volume", "issue", "number", "journal", "booktitle", "isbn"],
        "techreport": ["volume", "issue", "number", "journal", "booktitle", "isbn"],
        "unpublished": ["volume", "issue", "number", "journal", "booktitle", "isbn"],
        "online": ["journal", "booktitle", "isbn"],
        "misc": ["journal", "booktitle", "isbn"],
    }
    """Fields considered inconsistent with the respective ENTRYTYPE"""

    msg = "inconsistent-with-entrytype"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the completeness checks"""

        if record.data["ENTRYTYPE"] not in self.record_field_inconsistencies:
            return

        inconsistent_fields = self.record_field_inconsistencies[
            record.data["ENTRYTYPE"]
        ]
        for key in record.data:
            if key in inconsistent_fields:
                record.add_masterdata_provenance_note(key=key, note=self.msg)
            else:
                record.remove_masterdata_provenance_note(key=key, note=self.msg)


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(InconsistentWithEntrytypeChecker(quality_model))



#! /usr/bin/env python
"""Creation of a PRISMA chart as part of the data operations"""
from __future__ import annotations

import os
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import List
from typing import TYPE_CHECKING

import docker
import pandas as pd
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin
from docker.errors import DockerException

import colrev.env.package_manager
import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.data


@zope.interface.implementer(colrev.env.package_manager.DataPackageEndpointInterface)
@dataclass
class PRISMA(JsonSchemaMixin):
    """Create a PRISMA diagram"""

    ci_supported: bool = False

    @dataclass
    class PRISMASettings(colrev.env.package_manager.DefaultSettings, JsonSchemaMixin):
        """PRISMA settings"""

        endpoint: str
        version: str
        diagram_path: List[Path] = field(default_factory=lambda: [Path("PRISMA.png")])

    settings_class = PRISMASettings

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        # Set default values (if necessary)
        if "version" not in settings:
            settings["version"] = "0.1"

        if "diagram_path" in settings:
            settings["diagram_path"] = [Path(path) for path in settings["diagram_path"]]
        else:
            settings["diagram_path"] = [Path("PRISMA.png")]

        self.settings = self.settings_class.load_settings(data=settings)

        self.csv_path = data_operation.review_manager.output_dir / Path("PRISMA.csv")
        self.data_operation = data_operation

        self.settings.diagram_path = [
            data_operation.review_manager.output_dir / path
            for path in self.settings.diagram_path
        ]

        if not data_operation.review_manager.in_ci_environment():
            self.prisma_image = "colrev/prisma:latest"
            data_operation.review_manager.environment_manager.build_docker_image(
                imagename=self.prisma_image
            )

    # pylint: disable=unused-argument
    @classmethod
    def add_endpoint(cls, operation: colrev.ops.data.Data, params: str) -> None:
        """Add as an endpoint"""

        add_source = {
            "endpoint": "colrev.prisma",
            "version": "0.1",
            "diagram_path": ["PRISMA.png"],
        }
        operation.review_manager.settings.data.data_package_endpoints.append(add_source)

    def __export_csv(
        self, data_operation: colrev.ops.data.Data, silent_mode: bool
    ) -> None:
        csv_resource_path = Path("template/") / Path("prisma/PRISMA.csv")
        self.csv_path.parent.mkdir(exist_ok=True, parents=True)

        if self.csv_path.is_file():
            os.remove(self.csv_path)
        colrev.env.utils.retrieve_package_file(
            template_file=csv_resource_path, target=self.csv_path
        )

        status_stats = data_operation.review_manager.get_status_stats()

        prisma_data = pd.read_csv(self.csv_path)
        prisma_data["ind"] = prisma_data["data"]
        prisma_data.set_index("ind", inplace=True)
        prisma_data.loc["database_results", "n"] = status_stats.overall.md_retrieved
        prisma_data.loc[
            "duplicates", "n"
        ] = status_stats.currently.md_duplicates_removed
        prisma_data.loc["records_screened", "n"] = status_stats.overall.rev_prescreen
        prisma_data.loc[
            "records_excluded", "n"
        ] = status_stats.overall.rev_prescreen_excluded
        if status_stats.currently.exclusion:
            prisma_data.loc["dbr_excluded", "n"] = "; ".join(
                f"{key}, {val}" for key, val in status_stats.currently.exclusion.items()
            )
        else:
            prisma_data.loc[
                "dbr_excluded", "n"
            ] = f"Overall, {status_stats.overall.rev_excluded}"

        prisma_data.loc["dbr_assessed", "n"] = status_stats.overall.rev_screen
        prisma_data.loc["new_studies", "n"] = status_stats.overall.rev_included
        prisma_data.loc[
            "dbr_notretrieved_reports", "n"
        ] = status_stats.overall.pdf_not_available
        prisma_data.loc[
            "dbr_sought_reports", "n"
        ] = status_stats.overall.rev_prescreen_included

        prisma_data.to_csv(self.csv_path, index=False)
        data_operation.review_manager.logger.debug(f"Exported {self.csv_path}")

        if not status_stats.completeness_condition and not silent_mode:
            data_operation.review_manager.logger.info("Review not (yet) complete")

    def __export_diagram(
        self, data_operation: colrev.ops.data.Data, silent_mode: bool
    ) -> None:
        if not self.csv_path.is_file():
            data_operation.review_manager.logger.error(
                "File %s does not exist.", self.csv_path
            )
            data_operation.review_manager.logger.info(
                "Complete processing and use colrev data"
            )
            return

        csv_relative_path = self.csv_path.relative_to(
            data_operation.review_manager.path
        )

        if not silent_mode:
            data_operation.review_manager.logger.info("Create PRISMA diagram")

        for diagram_path in self.settings.diagram_path:
            diagram_relative_path = diagram_path.relative_to(
                data_operation.review_manager.path
            )

            script = (
                "Rscript "
                + "/prisma.R "
                + f"/data/{csv_relative_path} "
                + f"/data/{diagram_relative_path}"
            )
            # Users can place a custom script in src/prisma.R
            if (data_operation.review_manager.path / Path("src/prisma.R")).is_file():
                script = (
                    "Rscript "
                    + "-e \"source('/data/src/prisma.R')\" "
                    + f"/data/{csv_relative_path} "
                    + f"/data/{diagram_relative_path}"
                )

            self.__call_docker_build_process(
                data_operation=data_operation, script=script
            )
            csv_relative_path.unlink()

    def __call_docker_build_process(
        self, *, data_operation: colrev.ops.data.Data, script: str
    ) -> None:
        try:
            uid = os.stat(data_operation.review_manager.settings_path).st_uid
            gid = os.stat(data_operation.review_manager.settings_path).st_gid
            user = f"{uid}:{gid}"

            client = docker.from_env()

            msg = f"Running docker container created from image {self.prisma_image}"
            data_operation.review_manager.report_logger.info(msg)

            client.containers.run(
                image=self.prisma_image,
                command=script,
                user=user,
                volumes=[os.getcwd() + ":/data"],
            )
        except docker.errors.ImageNotFound:
            data_operation.review_manager.logger.error("Docker image not found")
        except docker.errors.ContainerError as exc:
            if "Temporary failure in name resolution" in str(exc):
                raise colrev_exceptions.ServiceNotAvailableException(
                    "prisma service failed"
                ) from exc
            raise exc
        except DockerException as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Docker service not available ({exc}). Please install/start Docker."
            ) from exc

    def update_data(
        self,
        data_operation: colrev.ops.data.Data,
        records: dict,  # pylint: disable=unused-argument
        synthesized_record_status_matrix: dict,  # pylint: disable=unused-argument
        silent_mode: bool,
    ) -> None:
        """Update the data/prisma diagram"""

        self.__export_csv(data_operation=data_operation, silent_mode=silent_mode)
        self.__export_diagram(data_operation=data_operation, silent_mode=silent_mode)

    def update_record_status_matrix(
        self,
        data_operation: colrev.ops.data.Data,  # pylint: disable=unused-argument
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record_status_matrix"""

        # Note : automatically set all to True / synthesized
        for syn_id in list(synthesized_record_status_matrix.keys()):
            synthesized_record_status_matrix[syn_id][endpoint_identifier] = True

    def get_advice(
        self,
        review_manager: colrev.review_manager.ReviewManager,  # pylint: disable=unused-argument
    ) -> dict:
        """Get advice on the next steps (for display in the colrev status)"""

        data_endpoint = "Data operation [prisma data endpoint]: "

        path_str = ",".join(
            [
                str(x.relative_to(review_manager.path))
                for x in self.settings.diagram_path
            ]
        )
        advice = {
            "msg": f"{data_endpoint}"
            + "\n    - The PRISMA diagram is created automatically "
            + f"({path_str})",
            "detailed_msg": "TODO",
        }
        return advice



#! /usr/bin/env python
"""Distribte records or PDFs to local CoLRev repositories."""
from __future__ import annotations

import os
import shutil
from pathlib import Path
from typing import TYPE_CHECKING

import colrev.operation
import colrev.settings

if TYPE_CHECKING:
    import colrev.review_manager


class Distribute(colrev.operation.Operation):
    """Distribute records to other local CoLRev projects"""

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        # pylint: disable=duplicate-code
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.check,
            notify_state_transition_operation=False,
        )
        self.review_manager = review_manager

    def get_environment_registry(self) -> list:
        """Get the environment registry (excluding curated_metadata)"""
        environment_manager = self.review_manager.get_environment_manager()
        return [
            x
            for x in environment_manager.local_repos()
            if "curated_metadata/" not in x["repo_source_path"]
        ]

    @colrev.operation.Operation.decorate()
    def main(self, *, path: Path, target: Path) -> None:
        """Distribute records to other CoLRev repositories (main entrypoint)"""

        # if no options are given, take the current path/repo
        # optional: target-repo-path
        # path_str: could also be a url
        # option: chdir (to target repo)?
        # file: copy or move?

        os.chdir(target)
        path = Path.cwd() / Path(path)

        if path.is_file():
            if path.suffix == ".bib":
                # gh_issue https://github.com/CoLRev-Environment/colrev/issues/69
                # append records (check duplicates/duplicate IDs)
                # if path already exists
                # should the following really rename the file?
                # or just get the updated filepath?

                path.rename(target / Path("data/search/local_import.bib"))
                input(path)

            if path.suffix == ".pdf":
                grobid_service = self.review_manager.get_grobid_service()

                grobid_service.start()

                tei = self.review_manager.get_tei(
                    pdf_path=path,
                )
                record = tei.get_metadata()

                target_pdf_path = target / "pdfs" / path.name
                target_pdf_path.parent.mkdir(parents=True, exist_ok=True)
                self.review_manager.logger.info(f"Copy PDF to {target_pdf_path}")
                shutil.copyfile(path, target_pdf_path)

                self.review_manager.logger.info(
                    f"append {self.review_manager.p_printer.pformat(record)} "
                    "to data/search/local_import.bib"
                )
                target_bib_file = target / Path("data/search/local_import.bib")
                self.review_manager.logger.info(f"target_bib_file: {target_bib_file}")
                if target_bib_file.is_file():
                    with open(target_bib_file, encoding="utf8") as target_bib:
                        import_records_dict = (
                            self.review_manager.dataset.load_records_dict(
                                load_str=target_bib.read()
                            )
                        )
                        import_records = list(import_records_dict.values())
                else:
                    import_records = []

                    new_source = colrev.settings.SearchSource(
                        endpoint="colrev.unknown_source",
                        filename=Path("search") / target_bib_file.name,
                        search_type=colrev.settings.SearchType.OTHER,
                        search_parameters={},
                        comment="",
                    )

                    self.review_manager.settings.sources.append(new_source)
                    self.review_manager.save_settings()

                if 0 != len(import_records):
                    record_id = int(
                        self.review_manager.dataset.get_next_id(
                            bib_file=target_bib_file
                        )
                    )

                record["ID"] = f"{record_id}".rjust(10, "0")
                record.update(file=str(target_pdf_path))
                import_records.append(record)

                import_records_dict = {r["ID"]: r for r in import_records}
                self.review_manager.dataset.save_records_dict_to_file(
                    records=import_records_dict, save_path=target_bib_file
                )

                self.review_manager.dataset.add_changes(path=target_bib_file)



#! /usr/bin/env python
"""Exclude collections as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

# pylint: disable=duplicate-code

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class ExcludeCollectionsPrep(JsonSchemaMixin):
    """Prepares records by excluding collection entries (e.g., proceedings)"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def prepare(
        self,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        record: colrev.record.PrepRecord,
    ) -> colrev.record.Record:
        """Prepare records by excluding collections (proceedings)"""

        if record.data["ENTRYTYPE"].lower() == "proceedings":
            record.prescreen_exclude(reason="collection/proceedings")

        return record



#! /usr/bin/env python
"""Template for a custom Screen PackageEndpoint"""
from __future__ import annotations

import random
from typing import TYPE_CHECKING

import zope.interface
from dacite import from_dict

import colrev.operation
import colrev.record

if TYPE_CHECKING:
    import colrev.screen.Screen

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.ScreenPackageEndpointInterface)
class CustomScreen:
    """Class for custom screen scripts"""

    def __init__(
        self,
        *,
        screen_operation: colrev.screen.Screen,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = from_dict(
            data_class=colrev.env.package_manager.DefaultSettings, data=settings
        )

    def run_screen(
        self, screen_operation: colrev.screen.Screen, records: dict, split: list
    ) -> dict:
        """Screen a record"""

        screen_data = screen_operation.get_data()
        screening_criteria = screen_operation.review_manager.settings.screen.criteria

        screening_criteria_available = bool(screening_criteria)

        for record_dict in screen_data["items"]:
            if len(split) > 0:
                if record_dict["ID"] not in split:
                    continue

            record = colrev.record.Record(data=record_dict)

            if random.random() < 0.5:  # nosec
                if screening_criteria_available:
                    # record criteria
                    pass
                screen_operation.screen(
                    record=record,
                    screen_inclusion=True,
                    screening_criteria="...",
                )

            else:
                if screening_criteria_available:
                    # record criteria
                    pass
                screen_operation.screen(
                    record=record,
                    screen_inclusion=False,
                    screening_criteria="...",
                )

        screen_operation.review_manager.dataset.save_records_dict(records=records)
        screen_operation.review_manager.dataset.add_record_changes()
        screen_operation.review_manager.create_commit(
            msg="Screen (random)", manual_author=False, script_call="colrev screen"
        )
        return records



#! /usr/bin/env python
"""Convenience functions to load bib files"""
from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import colrev.ops.load

# pylint: disable=too-few-public-methods


class NBIBLoader:

    """Loads nbib files"""

    def __init__(
        self,
        *,
        load_operation: colrev.ops.load.Load,
        source: colrev.settings.SearchSource,
    ):
        self.load_operation = load_operation
        self.source = source

    def load(
        self,
        *,
        source: colrev.settings.SearchSource,
    ) -> dict:
        """Converts ris entries it to bib records"""

        # pylint: disable=too-many-branches

        # Note : REFERENCE_TYPES and KEY_MAP are hard-coded (standard)
        # This function intentionally fails when the input does not comply
        # with this standard

        self.load_operation.ensure_append_only(file=self.source.filename)

        records = {}
        with open(source.filename, encoding="utf-8") as file:
            record = {}
            ind = 1
            for line in file:
                if line.startswith("TI "):
                    record["title"] = line[line.find(" - ") + 3 :].rstrip()

                elif line.startswith("JT "):
                    record["journal"] = line[line.find(" - ") + 3 :].rstrip()
                elif line.startswith("AB "):
                    record["abstract"] = line[line.find(" - ") + 3 :].rstrip()
                elif line.startswith("DP "):
                    record["year"] = line[line.find(" - ") + 3 :].rstrip()
                elif line.startswith("VI "):
                    record["volume"] = line[line.find(" - ") + 3 :].rstrip()
                elif line.startswith("IP "):
                    record["number"] = line[line.find(" - ") + 3 :].rstrip()

                elif line.startswith("PG "):
                    record["pages"] = line[line.find(" - ") + 3 :].rstrip()
                elif line.startswith("PT "):
                    if "Journal Articles" in line:
                        record["ENTRYTYPE"] = "article"
                elif line.startswith("AU "):
                    if "author" in record:
                        record["author"] += (
                            " and " + line[line.find(" - ") + 3 :].rstrip()
                        )
                    else:
                        record["author"] = line[line.find(" - ") + 3 :].rstrip()

                elif line.startswith("OT "):
                    if "keywords" in record:
                        record["keywords"] += (
                            ", " + line[line.find(" - ") + 3 :].rstrip()
                        )
                    else:
                        record["keywords"] = line[line.find(" - ") + 3 :].rstrip()

                elif line.rstrip() == "":
                    if not record:
                        continue
                    record["ID"] = str(ind).rjust(6, "0")
                    ind += 1
                    if "ENTRYTYPE" not in record:
                        if "booktitle" in record:
                            record["ENTRYTYPE"] = "inproceedings"
                        else:
                            record["ENTRYTYPE"] = "misc"
                    records[record["ID"]] = deepcopy(record)
                    record = {}
                # else:
                #     print(line)

        return records



#! /usr/bin/env python
"""Checker for language-unknown."""
from __future__ import annotations

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class LanguageChecker:
    """The LanguageChecker"""

    msg = "language-unknown"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the language-unknown checks"""

        if record.data.get("title", "UNKNOWN") == "UNKNOWN":
            return

        if self.__language_unknown(record=record):
            record.add_masterdata_provenance_note(key="title", note=self.msg)

        else:
            record.remove_masterdata_provenance_note(key="title", note=self.msg)

    def __language_unknown(self, *, record: colrev.record.Record) -> bool:
        if "language" in record.data:
            return False
        return True


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(LanguageChecker(quality_model))



#! /usr/bin/env python
"""Retrieving DOIs from a papers website/url as a prep operation"""
from __future__ import annotations

import collections
import re
from dataclasses import dataclass
from sqlite3 import OperationalError
from typing import TYPE_CHECKING

import requests
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.search_sources.doi_org as doi_connector
import colrev.ops.search_sources
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class DOIFromURLsPrep(JsonSchemaMixin):
    """Prepares records by retrieving its DOI from the website (URL)"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "check with the developer"
    always_apply_changes = False

    # https://www.crossref.org/blog/dois-and-matching-regular-expressions/
    doi_regex = re.compile(r"10\.\d{4,9}/[-._;/:A-Za-z0-9]*")

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)
        self.same_record_type_required = (
            prep_operation.review_manager.settings.is_curated_masterdata_repo()
        )
        try:
            self.session = prep_operation.review_manager.get_cached_session()
        except OperationalError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="sqlite-requests-cache"
            ) from exc
        _, self.email = prep_operation.review_manager.get_committer()

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare the record by retrieving its DOI from the website (url) if available"""

        if ("url" not in record.data and "fulltext" not in record.data) or (
            "doi" in record.data
        ):
            return record

        try:
            url = record.data.get("url", record.data.get("fulltext", "NA"))
            headers = {"user-agent": f"{__name__}  " f"(mailto:{self.email})"}
            ret = self.session.request(
                "GET", url, headers=headers, timeout=prep_operation.timeout
            )
            ret.raise_for_status()
            res = re.findall(self.doi_regex, ret.text)
            if not res:
                return record

            if len(res) == 1:
                ret_dois = [(res[0], 1)]
            else:
                counter = collections.Counter(res)
                ret_dois = counter.most_common()

            if not ret_dois:
                return record

            doi, _ = ret_dois[0]

            retrieved_record_dict = {
                "doi": doi.upper(),
                "ID": record.data["ID"],
            }
            retrieved_record = colrev.record.PrepRecord(data=retrieved_record_dict)
            doi_connector.DOIConnector.retrieve_doi_metadata(
                review_manager=prep_operation.review_manager,
                record=retrieved_record,
                timeout=prep_operation.timeout,
            )

            similarity = colrev.record.PrepRecord.get_retrieval_similarity(
                record_original=record,
                retrieved_record_original=retrieved_record,
                same_record_type_required=self.same_record_type_required,
            )
            if similarity < prep_operation.retrieval_similarity:
                return record

            record.merge(merging_record=retrieved_record, default_source=url)

        except (
            requests.exceptions.RequestException,
            colrev_exceptions.InvalidMerge,
            colrev_exceptions.RecordNotParsableException,
        ):
            pass
        return record



#! /usr/bin/env python
"""SearchSource: backward search (based on PDFs and GROBID)"""
from __future__ import annotations

import json
import typing
from dataclasses import dataclass
from pathlib import Path

import requests
import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin
from thefuzz import fuzz

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.search_sources.crossref
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class BackwardSearchSource(JsonSchemaMixin):
    """Performs a backward search extracting references from PDFs using GROBID
    Scope: all included papers with colrev_status in (rev_included, rev_synthesized)
    """

    __api_url = "https://opencitations.net/index/coci/api/v1/references/"

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "bwsearch_ref"
    search_type = colrev.settings.SearchType.BACKWARD_SEARCH
    api_search_supported = True
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "PDF backward search"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/pdf_backward_search.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        if "min_intext_citations" not in settings["search_parameters"]:
            settings["search_parameters"]["min_intext_citations"] = 3

        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        # Do not run in continuous-integration environment
        if not source_operation.review_manager.in_ci_environment():
            self.grobid_service = source_operation.review_manager.get_grobid_service()
            self.grobid_service.start()

        self.review_manager = source_operation.review_manager

        self.crossref_connector = (
            colrev.ops.built_in.search_sources.crossref.CrossrefSearchSource(
                source_operation=source_operation
            )
        )
        self.__etiquette = self.crossref_connector.get_etiquette(
            review_manager=self.review_manager
        )

    @classmethod
    def get_default_source(cls) -> colrev.settings.SearchSource:
        """Get the default SearchSource settings"""

        return colrev.settings.SearchSource(
            endpoint="colrev.pdf_backward_search",
            filename=Path("data/search/pdf_backward_search.bib"),
            search_type=colrev.settings.SearchType.BACKWARD_SEARCH,
            search_parameters={
                "scope": {"colrev_status": "rev_included|rev_synthesized"},
                "min_intext_citations": 3,
            },
            comment="",
        )

    def __validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""
        source = self.search_source
        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        if "scope" not in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "Scope required in the search_parameters"
            )

        if source.search_parameters["scope"].get("file", "") == "paper.md":
            pass
        else:
            if (
                source.search_parameters["scope"]["colrev_status"]
                != "rev_included|rev_synthesized"
            ):
                raise colrev_exceptions.InvalidQueryException(
                    "search_parameters/scope/colrev_status must be rev_included|rev_synthesized"
                )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def __bw_search_condition(self, *, record: dict) -> bool:
        # rev_included/rev_synthesized required, but record not in rev_included/rev_synthesized
        if (
            "colrev_status" in self.search_source.search_parameters["scope"]
            and self.search_source.search_parameters["scope"]["colrev_status"]
            == "rev_included|rev_synthesized"
            and record["colrev_status"]
            not in [
                colrev.record.RecordState.rev_included,
                colrev.record.RecordState.rev_synthesized,
            ]
        ):
            return False

        # Note: this is for peer_reviews
        if "file" in self.search_source.search_parameters["scope"]:
            if (
                self.search_source.search_parameters["scope"]["file"] == "paper.pdf"
            ) and "data/pdfs/paper.pdf" != record.get("file", ""):
                return False

        if not (self.review_manager.path / Path(record["file"])).is_file():
            self.review_manager.logger.error(f'File not found for {record["ID"]}')
            return False

        return True

    def __get_reference_records(self, *, record_dict: dict) -> list:
        references = []

        url = f"{self.__api_url}{record_dict['doi']}"
        # headers = {"authorization": "YOUR-OPENCITATIONS-ACCESS-TOKEN"}
        headers: typing.Dict[str, str] = {}
        ret = requests.get(url, headers=headers, timeout=300)
        try:
            items = json.loads(ret.text)

            for doi in [x["cited"] for x in items]:
                try:
                    retrieved_record = self.crossref_connector.query_doi(
                        doi=doi, etiquette=self.__etiquette
                    )
                    # if not crossref_query_return:
                    #     raise colrev_exceptions.RecordNotFoundInPrepSourceException()
                    retrieved_record.data["ID"] = retrieved_record.data["doi"]
                    references.append(retrieved_record.data)
                except (
                    colrev_exceptions.RecordNotFoundInPrepSourceException,
                    colrev_exceptions.RecordNotParsableException,
                ):
                    pass
        except json.decoder.JSONDecodeError:
            self.review_manager.logger.info(
                f"Error retrieving citations from Opencitations for {record_dict['ID']}"
            )

        return references

    def __get_similarity(
        self, *, record: colrev.record.Record, retrieved_record_dict: dict
    ) -> float:
        title_similarity = fuzz.partial_ratio(
            retrieved_record_dict.get("title", "NA").lower(),
            record.data.get("title", "").lower(),
        )
        container_similarity = fuzz.partial_ratio(
            colrev.record.PrepRecord(data=retrieved_record_dict)
            .get_container_title()
            .lower(),
            record.get_container_title().lower(),
        )
        weights = [0.6, 0.4]
        similarities = [title_similarity, container_similarity]

        similarity = sum(similarities[g] * weights[g] for g in range(len(similarities)))
        return similarity

    def __complement_with_open_citations_data(
        self,
        *,
        pdf_backward_search_feed: colrev.ops.search_feed.GeneralOriginFeed,
        records: dict,
    ) -> None:
        self.review_manager.logger.info("Comparing records with open-citations data")
        for parent_record_id in {
            x["bwsearch_ref"] for x in pdf_backward_search_feed.feed_records.values()
        }:
            parent_record = records[
                parent_record_id[: parent_record_id.find("_backward_search_")]
            ]

            if "doi" not in parent_record:
                continue

            backward_references = self.__get_reference_records(
                record_dict=parent_record
            )
            updated = 0
            overall = 0
            for feed_record_dict in pdf_backward_search_feed.feed_records.values():
                if feed_record_dict["bwsearch_ref"] != parent_record_id:
                    continue
                overall += 1
                feed_record = colrev.record.Record(data=feed_record_dict)
                max_similarity = 0.0
                pos = -1
                for i, backward_reference in enumerate(backward_references):
                    similarity = self.__get_similarity(
                        record=feed_record, retrieved_record_dict=backward_reference
                    )
                    if similarity > max_similarity:
                        pos = i
                        max_similarity = similarity
                if max_similarity > 0.9:
                    feed_record.data.update(**backward_references[pos - 1])
                    # del backward_references[pos-1]
                    updated += 1
            self.review_manager.logger.info(
                f" updated {updated}/{overall} records cited by {parent_record_id}"
            )

    def __run_backward_search_on_pdf(
        self,
        *,
        record: dict,
        pdf_backward_search_feed: colrev.ops.search_feed.GeneralOriginFeed,
        search_operation: colrev.ops.search.Search,
        records: dict,
        rerun: bool,
    ) -> None:
        # Note: IDs generated by GROBID for cited references
        # may change across grobid versions
        # -> challenge for key-handling/updating searches...

        self.review_manager.logger.info(f' run backward search for {record["ID"]}')

        pdf_path = self.review_manager.path / Path(record["file"])
        tei = self.review_manager.get_tei(
            pdf_path=pdf_path,
        )

        new_records = tei.get_bibliography(
            min_intext_citations=self.search_source.search_parameters[
                "min_intext_citations"
            ]
        )

        for new_record in new_records:
            if "tei_id" in new_record:
                del new_record["tei_id"]
            new_record["bwsearch_ref"] = (
                record["ID"] + "_backward_search_" + new_record["ID"]
            )
            try:
                pdf_backward_search_feed.set_id(record_dict=new_record)
            except colrev_exceptions.NotFeedIdentifiableException:
                continue

            prev_record_dict_version = {}
            if new_record["ID"] in pdf_backward_search_feed.feed_records:
                prev_record_dict_version = pdf_backward_search_feed.feed_records[
                    new_record["ID"]
                ]

            added = pdf_backward_search_feed.add_record(
                record=colrev.record.Record(data=new_record),
            )

            if not added and rerun:
                # Note : only re-index/update
                pdf_backward_search_feed.update_existing_record(
                    records=records,
                    record_dict=new_record,
                    prev_record_dict_version=prev_record_dict_version,
                    source=self.search_source,
                    update_time_variant_fields=rerun,
                )

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of PDFs (backward search based on GROBID)"""

        self.__validate_source()

        # Do not run in continuous-integration environment
        if search_operation.review_manager.in_ci_environment():
            return

        records = search_operation.review_manager.dataset.load_records_dict()

        if not records:
            search_operation.review_manager.logger.info(
                "No records imported. Cannot run backward search yet."
            )
            return

        search_operation.review_manager.logger.info(
            "Set min_intext_citations="
            f"{self.search_source.search_parameters['min_intext_citations']}"
        )

        pdf_backward_search_feed = self.search_source.get_feed(
            review_manager=search_operation.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )
        for record in records.values():
            try:
                if not self.__bw_search_condition(record=record):
                    continue

                self.__run_backward_search_on_pdf(
                    record=record,
                    pdf_backward_search_feed=pdf_backward_search_feed,
                    search_operation=search_operation,
                    records=records,
                    rerun=rerun,
                )

            except colrev_exceptions.TEIException:
                search_operation.review_manager.logger.info("Eror accessing TEI")

        self.__complement_with_open_citations_data(
            pdf_backward_search_feed=pdf_backward_search_feed, records=records
        )

        pdf_backward_search_feed.print_post_run_search_infos(
            records=records,
        )
        pdf_backward_search_feed.save_feed_file()

        if search_operation.review_manager.dataset.has_changes():
            search_operation.review_manager.create_commit(
                msg="Backward search", script_call="colrev search"
            )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for PDF backward searches (GROBID)"""

        result = {"confidence": 0.0}
        if str(filename).endswith("_ref_list.pdf"):
            result["confidence"] = 1.0
            return result
        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        if params is None:
            params = "min_intext_citations=3"

        if params == "default":
            add_source = cls.get_default_source()
            operation.review_manager.settings.sources.append(add_source)
            return

        if params.startswith("min_intext_citations="):
            add_source = cls.get_default_source()
            min_intext_citations = params.replace("min_intext_citations=", "")
            assert min_intext_citations.isdigit()
            add_source.search_parameters["min_intext_citations"] = int(
                min_intext_citations
            )
            operation.review_manager.settings.sources.append(add_source)
            return

        raise colrev_exceptions.PackageParameterError(
            f"Cannot add backward_search endpoint with query {params}"
        )

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def prepare(
        self, record: colrev.record.PrepRecord, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for PDF backward searches (GROBID)"""

        record.format_if_mostly_upper(key="title", case="sentence")
        record.format_if_mostly_upper(key="journal", case="title")
        record.format_if_mostly_upper(key="booktitle", case="title")
        record.format_if_mostly_upper(key="author", case="title")

        if (
            "multimedia appendix"
            in record.data.get("title", "").lower()
            + record.data.get("journal", "").lower()
        ):
            record.prescreen_exclude(reason="grobid-error")

        if record.data["ENTRYTYPE"] == "misc" and "publisher" in record.data:
            record.data["ENTRYTYPE"] = "book"

        return record



#! /usr/bin/env python
"""CoLRev prep_man operation: Prepare metadata manually."""
from __future__ import annotations

import csv
from pathlib import Path

import pandas as pd
from tqdm import tqdm

import colrev.env.language_service
import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.record


class PrepMan(colrev.operation.Operation):
    """Prepare records manually (metadata)"""

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.prep_man,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.verbose = True

    def __get_crosstab_df(self) -> pd.DataFrame:
        # pylint: disable=too-many-branches
        # pylint: disable=duplicate-code

        records = self.review_manager.dataset.load_records_dict()

        self.review_manager.logger.info("Calculate statistics")
        stats: dict = {"ENTRYTYPE": {}}
        overall_types: dict = {"ENTRYTYPE": {}}
        prep_man_hints, origins, crosstab = [], [], []
        for record_dict in records.values():
            if colrev.record.RecordState.md_imported != record_dict["colrev_status"]:
                if record_dict["ENTRYTYPE"] in overall_types["ENTRYTYPE"]:
                    overall_types["ENTRYTYPE"][record_dict["ENTRYTYPE"]] = (
                        overall_types["ENTRYTYPE"][record_dict["ENTRYTYPE"]] + 1
                    )
                else:
                    overall_types["ENTRYTYPE"][record_dict["ENTRYTYPE"]] = 1

            if (
                colrev.record.RecordState.md_needs_manual_preparation
                != record_dict["colrev_status"]
            ):
                continue

            if record_dict["ENTRYTYPE"] in stats["ENTRYTYPE"]:
                stats["ENTRYTYPE"][record_dict["ENTRYTYPE"]] = (
                    stats["ENTRYTYPE"][record_dict["ENTRYTYPE"]] + 1
                )
            else:
                stats["ENTRYTYPE"][record_dict["ENTRYTYPE"]] = 1

            if "colrev_masterdata_provenance" in record_dict:
                record = colrev.record.Record(data=record_dict)
                prov_d = record.data["colrev_masterdata_provenance"]
                hints = []
                for key, value in prov_d.items():
                    if value["note"] != "":
                        hints.append(f'{key} - {value["note"]}')

                prep_man_hints.append([hint.lstrip() for hint in hints])
                for hint in hints:
                    if "change-score" in hint:
                        continue
                    # Note: if something causes the needs_manual_preparation
                    # it is caused by all colrev_origins
                    for orig in record_dict.get("colrev_origin", ["NA"]):
                        crosstab.append([orig[: orig.rfind("/")], hint.lstrip()])

            origins.append(
                [x[: x.rfind("/")] for x in record_dict.get("colrev_origin", ["NA"])]
            )

        print("Entry type statistics overall:")
        self.review_manager.p_printer.pprint(overall_types["ENTRYTYPE"])

        print("Entry type statistics (needs_manual_preparation):")
        self.review_manager.p_printer.pprint(stats["ENTRYTYPE"])

        return pd.DataFrame(crosstab, columns=["colrev_origin", "hint"])

    def prep_man_langs(self) -> None:
        """Add missing language fields based on spreadsheets"""

        lang_prep_csv_path = self.review_manager.prep_dir / Path(
            "missing_lang_recs_df.csv"
        )
        lang_prep_csv_path.parent.mkdir(exist_ok=True, parents=True)

        records = self.review_manager.dataset.load_records_dict()

        if not lang_prep_csv_path.is_file():
            language_service = colrev.env.language_service.LanguageService()

            missing_lang_recs = []
            self.review_manager.logger.info(
                "Calculate most likely languages for records without language field"
            )
            for record in tqdm(records.values()):
                if "title" not in record:
                    continue
                if "language" not in record:
                    confidence_values = (
                        language_service.compute_language_confidence_values(
                            text=record["title"]
                        )
                    )
                    predicted_language, conf = confidence_values.pop(0)

                    lang_rec = {
                        "ID": record["ID"],
                        "title": record["title"],
                        "most_likely_language": predicted_language,
                        "confidence": conf,
                    }
                    missing_lang_recs.append(lang_rec)

            missing_lang_recs_df = pd.DataFrame(missing_lang_recs)
            missing_lang_recs_df.to_csv(
                lang_prep_csv_path, index=False, quoting=csv.QUOTE_ALL
            )
            self.review_manager.logger.info(f"Exported table to {lang_prep_csv_path}")
            self.review_manager.logger.info(
                "Update the language column and rerun colrev prep-man -l"
            )

        else:
            self.review_manager.logger.info(
                f"Import language fields from {lang_prep_csv_path}"
            )
            languages_df = pd.read_csv(lang_prep_csv_path)
            language_records = languages_df.to_dict("records")
            for language_record in language_records:
                if language_record["most_likely_language"] == "":
                    continue
                if language_record["ID"] not in records:
                    # warn
                    continue
                record_dict = records[language_record["ID"]]
                record = colrev.record.Record(data=record_dict)
                record.update_field(
                    key="language",
                    value=language_record["most_likely_language"],
                    source="LanguageDetector/Manual",
                    note="",
                )
                if "language of title not in [eng]" == record.data.get(
                    "prescreen_exclusion", ""
                ):
                    record.remove_field(key="prescreen_exclusion")
                    if "title" in record.data["colrev_masterdata_provenance"]:
                        record.data["colrev_masterdata_provenance"]["title"][
                            "note"
                        ] = ""
                    if (
                        record.data["colrev_status"]
                        == colrev.record.RecordState.md_needs_manual_preparation
                    ):
                        # by resetting to md_imported,
                        # the prescreen-exclusion based on languages will be reapplied.
                        record.set_status(
                            target_state=colrev.record.RecordState.md_imported
                        )

            self.review_manager.dataset.save_records_dict(records=records)

    def prep_man_stats(self) -> None:
        """Print statistics on prep_man"""
        # pylint: disable=duplicate-code

        crosstab_df = self.__get_crosstab_df()

        if crosstab_df.empty:
            print("No records to prepare manually.")
        else:
            # pylint: disable=duplicate-code
            tabulated = pd.pivot_table(
                crosstab_df[["colrev_origin", "hint"]],
                index=["colrev_origin"],
                columns=["hint"],
                aggfunc=len,
                fill_value=0,
                margins=True,
            )
            # .sort_index(axis='columns')
            tabulated.sort_values(by=["All"], ascending=False, inplace=True)
            # Transpose because we tend to have more error categories than search files.
            tabulated = tabulated.transpose()
            print(tabulated)
            self.review_manager.logger.info(
                "Writing data to file: manual_preparation_statistics.csv"
            )
            tabulated.to_csv("manual_preparation_statistics.csv")

    def get_data(self) -> dict:
        """Get the data for prep-man"""
        # pylint: disable=duplicate-code

        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())
        nr_tasks = len(
            [
                x
                for x in record_header_list
                if colrev.record.RecordState.md_needs_manual_preparation
                == x["colrev_status"]
            ]
        )

        all_ids = [x["ID"] for x in record_header_list]

        pad = min((max(len(x["ID"]) for x in record_header_list) + 2), 35)

        items = self.review_manager.dataset.read_next_record(
            conditions=[
                {"colrev_status": colrev.record.RecordState.md_needs_manual_preparation}
            ]
        )

        md_prep_man_data = {
            "nr_tasks": nr_tasks,
            "items": items,
            "all_ids": all_ids,
            "PAD": pad,
        }
        self.review_manager.logger.debug(
            self.review_manager.p_printer.pformat(md_prep_man_data)
        )
        return md_prep_man_data

    def set_data(self, *, record_dict: dict) -> None:
        """Set data in the prep_man operation"""

        record = colrev.record.PrepRecord(data=record_dict)
        record.set_masterdata_complete(
            source="prep_man",
            masterdata_repository=self.review_manager.settings.is_curated_repo(),
        )
        record.set_masterdata_consistent()
        # record.set_fields_complete()
        record.set_status(target_state=colrev.record.RecordState.md_prepared)
        record_dict = record.get_data()

        self.review_manager.dataset.save_records_dict(
            records={record_dict["ID"]: record_dict}, partial=True
        )
        self.review_manager.dataset.add_record_changes()

    @colrev.operation.Operation.decorate()
    def main(self) -> None:
        """Manually prepare records (main entrypoint)"""

        if self.review_manager.in_ci_environment():
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="colrev prep-man",
                detailed_trace="prep-man not available in ci environment",
            )

        records = self.review_manager.dataset.load_records_dict()

        package_manager = self.review_manager.get_package_manager()

        for (
            prep_man_package_endpoint
        ) in self.review_manager.settings.prep.prep_man_package_endpoints:
            endpoint_dict = package_manager.load_packages(
                package_type=colrev.env.package_manager.PackageEndpointType.prep_man,
                selected_packages=self.review_manager.settings.prep.prep_man_package_endpoints,
                operation=self,
            )
            endpoint = endpoint_dict[prep_man_package_endpoint["endpoint"]]
            records = endpoint.prepare_manual(self, records)  # type: ignore



#!/usr/bin/env python3
"""Scripts to print the CoLRev status (cli)."""
from __future__ import annotations

import os
import subprocess  # nosec
from typing import TYPE_CHECKING

import colrev.record
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.ops.status

# pylint: disable=duplicate-code
keys = [
    "author",
    "title",
    "journal",
    "booktitle",
    "year",
    "volume",
    "number",
    "pages",
]


def __validate_dedupe(
    *,
    validate_operation: colrev.operation.Operation,
    validation_details: dict,
    threshold: float,  # pylint: disable=unused-argument
) -> None:
    dedupe_operation = validate_operation.review_manager.get_dedupe_operation()

    for validation_item in validation_details:
        print(validation_item["change_score"])
        colrev.record.Record.print_diff_pair(
            record_pair=[
                validation_item["prior_record_a"],
                validation_item["prior_record_b"],
            ],
            keys=keys,
        )

        user_selection = input("Validate [y,n,q for yes, no (undo), or quit]?")

        if user_selection == "n":
            dedupe_operation.unmerge_records(
                current_record_ids=validation_item["record"]["ID"]
            )

        if user_selection == "q":
            break
        if user_selection == "y":
            continue


def __validate_prep(
    *,
    validate_operation: colrev.operation.Operation,
    validation_details: list,
    threshold: float,
) -> None:
    # Note : for testing:
    # prescreen_excluded_to_validate = [
    # e for e in validation_details if not e["prescreen_exclusion_mark"]
    # ]
    prescreen_excluded_to_validate = [
        e for e in validation_details if e["prescreen_exclusion_mark"]
    ]
    print("Prescreen excluded:")
    for i, validation_detail in enumerate(prescreen_excluded_to_validate):
        print(i)
        colrev.record.Record(
            data=validation_detail["record_dict"]
        ).print_citation_format()

    displayed = False
    for validation_element in validation_details:
        if validation_element["change_score"] < threshold:
            continue
        displayed = True
        # Escape sequence to clear terminal output for each new comparison
        os.system("cls" if os.name == "nt" else "clear")
        if (
            validation_element["prior_record_dict"]["ID"]
            == validation_element["record_dict"]["ID"]
        ):
            print(
                f"difference: {str(round(validation_element['change_score'], 4))} "
                f"record {validation_element['prior_record_dict']['ID']}"
            )
        else:
            print(
                f"difference: {str(round(validation_element['change_score'], 4))} "
                f"record {validation_element['prior_record_dict']['ID']} - "
                f"{validation_element['record_dict']['ID']}"
            )

        colrev.record.Record.print_diff_pair(
            record_pair=[
                validation_element["prior_record_dict"],
                validation_element["record_dict"],
            ],
            keys=keys,
        )

        user_selection = input("Validate [y,n,q for yes, no (undo), or quit]?")

        if user_selection == "n":
            validate_operation.review_manager.dataset.save_records_dict(
                records={
                    validation_element["prior_record_dict"]["ID"]: validation_element[
                        "prior_record_dict"
                    ]
                },
                partial=True,
            )

        if user_selection == "q":
            break
        if user_selection == "y":
            continue

    if not displayed:
        validate_operation.review_manager.logger.info(
            "No preparation changes above threshold"
        )


def validate(
    *,
    validate_operation: colrev.operation.Operation,
    validation_details: dict,
    threshold: float,
) -> None:
    """Validate details in the cli"""

    for key, details in validation_details.items():
        if key == "prep":
            __validate_prep(
                validate_operation=validate_operation,
                validation_details=details,
                threshold=threshold,
            )
        elif key == "dedupe":
            __validate_dedupe(
                validate_operation=validate_operation,
                validation_details=details,
                threshold=threshold,
            )
        elif key == "properties":
            validate_operation.review_manager.logger.info(
                " Traceability of records".ljust(32, " ")
                + str(details["record_traceability"])
            )
            validate_operation.review_manager.logger.info(
                " Consistency (based on hooks)".ljust(32, " ")
                + str(details["consistency"])
            )
            validate_operation.review_manager.logger.info(
                " Completeness of iteration".ljust(32, " ")
                + str(details["completeness"])
            )
        elif key == "contributor_commits":
            validate_operation.review_manager.logger.info(
                "Showing commits in which the contributor was involved as the author or committer."
            )

            print()
            print("Commits to validate:")
            print()
            for item in details:
                for _, item_values in item.items():
                    print(item_values["msg"])
                    print(f"  date      {item_values['date']}")
                    print(
                        f"  author    {item_values['author']} ({item_values['author_email']})"
                    )
                    print(
                        f"  committer {item_values['committer']} ({item_values['committer_email']})"
                    )
                    print(f"  {colors.ORANGE}{item_values['validate']}{colors.END}")

            print()
        elif key == "general":
            validate_operation.review_manager.logger.info("Start general validation")
            validate_operation.review_manager.logger.info(
                "Next, an interface will open and "
                "display the changes introduced in the selected commit."
            )
            validate_operation.review_manager.logger.info(
                "To undo minor changes, edit the corresponding files directly and run "
                f"{colors.ORANGE}git add FILENAME && "
                f"git commit -m 'DESCRIPTION OF CHANGES UNDONE'{colors.END}"
            )
            validate_operation.review_manager.logger.info(
                "To undo all changes introduced in a commit, run "
                f"{colors.ORANGE}git revert COMMIT_ID{colors.END}"
            )
            input("Enter to continue")
            if "commit_relative" in details:
                subprocess.run(  # nosec
                    ["gitk", f"--select-commit={details['commit_relative']}"],
                    check=False,
                )
            else:
                subprocess.run(["gitk"], check=False)  # nosec

        else:
            print("Not yet implemented")
            print(validation_details)

    if validate_operation.review_manager.dataset.records_changed():
        validate_operation.review_manager.create_commit(msg="validate")



#! /usr/bin/env python
"""SearchSource: GoogleScholar"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.load_utils_bib
import colrev.ops.search
import colrev.record


# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class GoogleScholarSearchSource(JsonSchemaMixin):
    """SearchSource for GoogleScholar"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "url"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "GoogleScholar"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/google_scholar.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for GoogleScholar"""

        result = {"confidence": 0.0}
        if data.count("https://scholar.google.com/scholar?q=relat") > 0.9 * data.count(
            "\n@"
        ):
            result["confidence"] = 1.0
            return result

        if data.count("{pop0") > 0.9 * data.count("\n@"):
            result["confidence"] = 1.0
            return result

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        raise NotImplementedError

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of GoogleScholar"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for GoogleScholar"""
        if "note" in record.data:
            if (
                "cites: https://scholar.google.com/scholar?cites="
                in record.data["note"]
            ):
                note = record.data["note"]
                source = record.data["colrev_data_provenance"]["note"]["source"]
                record.rename_field(key="note", new_key="cited_by")
                record.update_field(
                    key="cited_by",
                    value=record.data["cited_by"][
                        : record.data["cited_by"].find(" cites: ")
                    ],
                    source="replace_link",
                )
                record.update_field(
                    key="cited_by_link",
                    value=note[note.find("cites: ") + 7 :],
                    append_edit=False,
                    source=source + "|extract-from-note",
                )
        if "abstract" in record.data:
            # Note: abstracts provided by GoogleScholar are very incomplete
            record.remove_field(key="abstract")

        return record



#! /usr/bin/env python
"""Exclude records based on language as a prep operation"""
from __future__ import annotations

import re
import statistics
from dataclasses import dataclass
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.language_service
import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class ExcludeLanguagesPrep(JsonSchemaMixin):
    """Prepares records by excluding ones that are not in the languages_to_include"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    def __init__(self, *, prep_operation: colrev.ops.prep.Prep, settings: dict) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

        # Note : the following objects have heavy memory footprints and should be
        # class (not object) properties to keep parallel processing as
        # efficient as possible (the object is passed to each thread)
        languages_to_include = ["eng"]
        # if not prep_operation.review_manager.in_ci_environment():
        self.language_service = colrev.env.language_service.LanguageService()

        prescreen_package_endpoints = (
            prep_operation.review_manager.settings.prescreen.prescreen_package_endpoints
        )
        if "scope_prescreen" in [s["endpoint"] for s in prescreen_package_endpoints]:
            for scope_prescreen in [
                s
                for s in prescreen_package_endpoints
                if "scope_prescreen" == s["endpoint"]
            ]:
                languages_to_include.extend(
                    scope_prescreen.get("LanguageScope", ["eng"])
                )
        self.language_service.validate_iso_639_3_language_codes(
            lang_code_list=languages_to_include
        )
        self.languages_to_include = list(set(languages_to_include))

    def __title_has_multiple_languages(self, *, title: str) -> bool:
        if "[" not in title:
            return False
        split_titles = [
            re.sub(r"\d", "", x.rstrip().rstrip("]")) for x in title.split("[")
        ]
        min_len = statistics.mean(len(x) for x in split_titles) - 20
        if all(len(x) > min_len for x in split_titles):
            return True
        return False

    def prepare(
        self,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        record: colrev.record.PrepRecord,
    ) -> colrev.record.Record:
        """Prepare the record by excluding records whose metadata is not in English"""

        # Note : other languages are not yet supported
        # because the dedupe does not yet support cross-language merges

        if record.data.get("title", "UNKNOWN") == "UNKNOWN":
            record.remove_field(key="language")
            return record

        if "language" in record.data:
            # Note: classification of non-english titles is not reliable.
            # Other languages should be checked in man-prep.
            # if "eng" != language:
            #     return record

            # if record.data["language"] not in self.languages_to_include:
            #     record.prescreen_exclude(
            #         reason=(
            #             "language of title not in "
            #             f"[{','.join(self.languages_to_include)}]"
            #         )
            #     )
            return record

        # To avoid misclassifications for short titles
        if len(record.data.get("title", "")) < 30:
            # If language not in record, add language
            # (always - needed in dedupe.)
            record.data["language"] = "eng"
            return record

        if not self.__title_has_multiple_languages(title=record.data.get("title", "")):
            language = self.language_service.compute_language(text=record.data["title"])
            # Note: classification of non-english titles is not reliable.
            # Other languages should be checked in man-prep.
            if "eng" != language:
                return record
            record.update_field(
                key="language",
                value=language,
                source="LanguageDetector",
                note="",
                append_edit=False,
            )
        else:
            # Deal with title fields containing titles in two or more languages
            split_titles = [
                x.rstrip().rstrip("]") for x in record.data["title"].split("[")
            ]
            for i, split_title in enumerate(split_titles):
                lang_split_title = self.language_service.compute_language(
                    text=split_title
                )
                if 0 == i:
                    record.update_field(
                        key="title",
                        value=split_title.rstrip(),
                        source="LanguageDetector_split",
                    )
                    record.update_field(
                        key="language",
                        value=lang_split_title,
                        source="LanguageDetector_split",
                    )
                else:
                    record.update_field(
                        key=f"title_{lang_split_title}",
                        value=split_title.rstrip("]"),
                        source="LanguageDetector_split",
                    )

        if record.data.get("language", "") == "":
            record.update_field(
                key="title",
                value=record.data.get("title", ""),
                source="LanguageDetector",
                note="language-not-found",
            )
            record.set_status(
                target_state=colrev.record.RecordState.md_needs_manual_preparation
            )
            return record

        if record.data.get("language", "") not in self.languages_to_include:
            record.remove_field(key="language")
            # record.prescreen_exclude(
            #     reason=f"language of title not in [{','.join(self.languages_to_include)}]"
            # )
            record.set_status(
                target_state=colrev.record.RecordState.md_needs_manual_preparation
            )
        return record



#!/usr/bin/env python3
"""Remove records, ... from CoLRev projects."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import colrev.env.utils
import colrev.operation

if TYPE_CHECKING:
    import colrev.review_manager


# pylint: disable=too-few-public-methods


class Remove(colrev.operation.Operation):
    """Remove records, ... from CoLRev projects."""

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.check,
            notify_state_transition_operation=False,
        )

    @colrev.operation.Operation.decorate()
    def remove_records(self, *, ids: str) -> None:
        """Remove records from CoLRev project."""

        records = self.review_manager.dataset.load_records_dict()

        for record_id in ids.split(","):
            if record_id in records:
                self.review_manager.logger.info(f" remove {record_id}")
                origins = records[record_id]["colrev_origin"]
                if "file" in records[record_id]:
                    print(f"manually remove file: {records[record_id]['file']}")
                del records[record_id]
                for origin in origins:
                    file, origin_id = origin.split("/")

                    filepath = self.review_manager.search_dir / Path(file)

                    origin_file_content = filepath.read_text()
                    origin_records = self.review_manager.dataset.load_records_dict(
                        load_str=origin_file_content
                    )
                    if origin_id in origin_records:
                        del origin_records[origin_id]
                    self.review_manager.dataset.save_records_dict_to_file(
                        records=origin_records, save_path=filepath
                    )
                    self.review_manager.dataset.add_changes(path=filepath)

        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()

        self.review_manager.create_commit(msg="Remove records", manual_author=False)



#! /usr/bin/env python
"""Removal of broken IDs as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class RemoveBrokenIDPrep(JsonSchemaMixin):
    """Prepares records by removing invalid IDs DOIs/ISBNs"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    # check_status: relies on crossref / openlibrary connectors!
    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare the record by removing broken IDs (invalid DOIs/ISBNs)"""

        if prep_operation.polish and not prep_operation.force_mode:
            return record

        if "doi" in record.data:
            if "doi" in record.data.get("colrev_masterdata_provenance", {}):
                if "doi-not-matching-pattern" in record.data[
                    "colrev_masterdata_provenance"
                ]["doi"]["note"].split(","):
                    record.remove_field(key="doi")

        if "isbn" in record.data:
            if "isbn" in record.data.get("colrev_masterdata_provenance", {}):
                if "isbn-not-matching-pattern" in record.data[
                    "colrev_masterdata_provenance"
                ]["isbn"]["note"].split(","):
                    record.remove_field(key="isbn")

        return record



#! /usr/bin/env python
"""Consolidation of metadata based on Europe PMC API as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.built_in.search_sources.europe_pmc as europe_pmc_connector
import colrev.ops.search_sources
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class EuropePMCMetadataPrep(JsonSchemaMixin):
    """Prepares records based on Europe PCM metadata"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "ask the publisher to correct the metadata"
    always_apply_changes = False

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare a record based on Europe PMC metadata"""

        # pylint: disable=invalid-name
        EuropePMCSearchSource = europe_pmc_connector.EuropePMCSearchSource(
            source_operation=prep_operation
        )
        EuropePMCSearchSource.get_masterdata(
            prep_operation=prep_operation, record=record
        )
        return record



#!/usr/bin/env python3
"""Scripts to print the CoLRev status (cli)."""
from __future__ import annotations

import sys
import typing
from time import sleep
from typing import TYPE_CHECKING

from tqdm import tqdm

import colrev.exceptions as colrev_exceptions
import colrev.ui_cli.cli_colors as colors
from colrev.exit_codes import ExitCodes

if TYPE_CHECKING:
    import colrev.ops.status


def print_review_instructions(review_instructions: dict) -> None:
    """Print the review instructions on cli"""

    print("Next operation")

    verbose = False

    # key_list = [list(x.keys()) for x in review_instructions]
    # keys = [item for sublist in key_list for item in sublist]
    # priority_item_set = "priority" in keys
    if not review_instructions:
        print(f"    {colors.GREEN}Review iteration completed{colors.END}")
        print(
            f"    {colors.ORANGE}To start the next iteration of the review, "
            f"add new search results (to data/search){colors.END}"
        )
        print()

    for review_instruction in review_instructions:
        # prioritize based on the order of instructions (most important first)
        # if priority_item_set and "priority" not in review_instruction.keys():
        #     continue

        if "info" in review_instruction:
            print("    " + review_instruction["info"])
        if "msg" in review_instruction:
            if "cmd" in review_instruction:
                if verbose:
                    print("    " + review_instruction["msg"] + ", i.e., use ")
                print(f'    {colors.ORANGE}{review_instruction["cmd"]}{colors.END}')
            else:
                print(f"    {colors.ORANGE}{review_instruction['msg']}{colors.END}")
        if "cmd_after" in review_instruction:
            print("    Then use " + review_instruction["cmd_after"])
        print()


def __print_collaboration_instructions_status(
    *, collaboration_instructions: dict
) -> None:
    if "status" not in collaboration_instructions:
        return
    if "title" in collaboration_instructions["status"]:
        title = collaboration_instructions["status"]["title"]
        if collaboration_instructions["status"].get("level", "NA") == "WARNING":
            print(f"  {colors.RED}{title}{colors.END}")
        elif collaboration_instructions["status"].get("level", "NA") == "SUCCESS":
            print(f"  {colors.GREEN}{title}{colors.END}")
        else:
            print("  " + title)
    if "msg" in collaboration_instructions["status"]:
        print(f'  {collaboration_instructions["status"]["msg"]}')


def __print_collaboration_instructions_items(
    *, collaboration_instructions: dict
) -> None:
    for item in collaboration_instructions["items"]:
        if "title" in item:
            if "level" in item:
                if item["level"] == "WARNING":
                    print(f'  {colors.RED}{item["title"]}{colors.END}')
                elif item["level"] == "SUCCESS":
                    print(f'  {colors.GREEN}{item["title"]}{colors.END}')
            else:
                print("  " + item["title"])

        if "msg" in item:
            print("  " + item["msg"])
        if "cmd_after" in item:
            print(f'  {item["cmd_after"]}')
        print()


def print_collaboration_instructions(
    *, status_operation: colrev.ops.status.Status, collaboration_instructions: dict
) -> None:
    """Print the collaboration instructions on cli"""

    if (
        not status_operation.review_manager.verbose_mode
        and collaboration_instructions["items"]
    ):
        if collaboration_instructions["items"][0]["title"] in [
            "Project not yet shared",
            "Up-to-date",
        ]:
            return

    print("Versioning and collaboration")
    __print_collaboration_instructions_status(
        collaboration_instructions=collaboration_instructions
    )
    __print_collaboration_instructions_items(
        collaboration_instructions=collaboration_instructions
    )


def print_environment_instructions(environment_instructions: dict) -> None:
    """Print the environment instructions on cli"""

    if not environment_instructions:
        return

    print("CoLRev environment\n")

    key_list = [list(x.keys()) for x in environment_instructions]
    keys = [item for sublist in key_list for item in sublist]
    priority_item_set = "priority" in keys

    for environment_instruction in environment_instructions:
        if priority_item_set and "priority" not in environment_instruction.keys():
            continue
        if "info" in environment_instruction:
            print("  " + environment_instruction["info"])
        if "msg" in environment_instruction:
            if "cmd" in environment_instruction:
                print("  " + environment_instruction["msg"] + "  i.e., use ")
                print(f'  {colors.ORANGE}{environment_instruction["cmd"]}{colors.END}')
            else:
                print("  " + environment_instruction["msg"])
        if "cmd_after" in environment_instruction:
            print("  Then use " + environment_instruction["cmd_after"])
        print()


def print_progress(*, total_atomic_steps: int, completed_steps: int) -> None:
    """Print the progress bar on cli"""

    # Prints the percentage of atomic processing tasks that have been completed
    # possible extension: estimate the number of manual tasks (making assumptions on
    # frequencies of man-prep, ...)?

    if total_atomic_steps != 0:
        current_percentage = int((completed_steps / total_atomic_steps) * 100)
    else:
        current_percentage = -1

    sleep_interval = 1.1 / max(current_percentage, 100)
    print()

    for i in tqdm(
        range(100),
        desc="    Progress:",
        bar_format="{desc} |{bar}|{percentage:.0f}%",
        ncols=40,
    ):
        sleep(sleep_interval)
        if current_percentage in [i, -1]:
            break


def print_project_status(status_operation: colrev.ops.status.Status) -> None:
    """Print the project status on cli"""

    # pylint: disable=too-many-statements
    # pylint: disable=too-many-branches

    print("Load records...", end="\r")
    sys.stdout.write("\033[K")

    ret_check: typing.Dict[str, typing.Any] = {"status": ExitCodes.SUCCESS}
    failure_items = []
    try:
        checker = status_operation.review_manager.get_checker()
        failure_items.extend(checker.check_repo_basics())
    except colrev_exceptions.RepoSetupError as exc:
        ret_check = {"status": ExitCodes.FAIL, "msg": exc}

    try:
        status_stats = status_operation.review_manager.get_status_stats(
            records=checker.records
        )
        status_report = status_operation.get_review_status_report(
            records=checker.records
        )
        print(status_report)

        if (
            not status_stats.completeness_condition
            and status_operation.review_manager.verbose_mode
        ):
            print_progress(
                total_atomic_steps=status_stats.atomic_steps,
                completed_steps=status_stats.completed_atomic_steps,
            )
        print("")

        advisor = status_operation.review_manager.get_advisor()
        instructions = advisor.get_instructions(status_stats=status_stats)
        print_review_instructions(instructions["review_instructions"])
        print_collaboration_instructions(
            status_operation=status_operation,
            collaboration_instructions=instructions["collaboration_instructions"],
        )
        print_environment_instructions(instructions["environment_instructions"])

    except colrev_exceptions.RepoSetupError as exc:
        print(f"Status failed ({exc})")

    # if status_operation.review_manager.verbose_mode:
    #     print("Checks")

    try:
        failure_items.extend(checker.check_repo_extended())
    except colrev_exceptions.RepoSetupError as exc:
        ret_check = {"status": ExitCodes.FAIL, "msg": exc}

    if failure_items:
        ret_check = {"status": ExitCodes.FAIL, "msg": "  " + "\n  ".join(failure_items)}

    # if (
    #     ExitCodes.SUCCESS == ret_check["status"]
    #     and status_operation.review_manager.verbose_mode
    # ):
    #     print(
    #         "  ReviewManager.check_repo()  ...  "
    #         f"{colors.GREEN}Everything ok.{colors.END}"
    #     )
    if ExitCodes.FAIL == ret_check["status"]:
        # print(f"  ReviewManager.check_repo()  ...  {colors.RED}FAIL{colors.END}")
        # print(f'{ret_check["msg"]}\n')
        return

    # To format:
    # status_operation.review_manager.dataset.save_records_dict(records=checker.records)

    # if (
    #     not status_operation.review_manager.in_virtualenv()
    #     and status_operation.review_manager.verbose_mode
    # ):
    #     print(
    #         f"\n  {colors.RED}WARNING{colors.END} running scripts outside of virtualenv"
    #     )
    #     print(
    #         "  For instructions to set up a virtual environment, run\n"
    #         f"  {colors.ORANGE}colrev show venv{colors.END}"
    #     )
    #     print()

    if status_operation.review_manager.verbose_mode:
        print(
            "Documentation: https://colrev.readthedocs.io/en/latest/manual/manual.html"
        )
    else:
        print(f"For more details: {colors.ORANGE}colrev status -v{colors.END}")



#!/usr/bin/env python3
"""Linter for CoLRev"""
from __future__ import annotations

from typing import TYPE_CHECKING

from astroid import nodes
from pylint import checkers
from pylint import interfaces
from pylint.checkers.utils import only_required_for_messages

if TYPE_CHECKING:
    from pylint.lint import PyLinter


class DirectStatusAssignmentChecker(checkers.BaseChecker):
    """DirectStatusAssignmentChecker"""

    __implements__ = interfaces.IAstroidChecker

    name = "direct-status-assign"

    msgs = {
        "W0931": (
            "Direct status assignment",
            "direct-status-assign",
            "Emitted when a colrev-status is directly assigned "
            + "(instead of using record.set_stautus)",
        ),
    }

    @only_required_for_messages("direct-status-assign")
    def visit_assign(self, node: nodes.Assign) -> None:
        """
        Detect direct assignment of colrev_status.
        """

        # if 706 != node.fromlineno:
        #   return

        # if not isinstance(node.targets[0], nodes.Dict):
        #     return
        if len(node.targets) != 1:
            return

        if not hasattr(node.targets[0], "slice"):
            return

        if not hasattr(node.targets[0].slice, "value"):
            return

        if "colrev_status" == node.targets[0].slice.value:
            # for variable, value in vars(node.targets[0].slice).items():
            #   if variable[:1] == '_':
            #       print(f"{variable}{value}")
            # print(node.targets[0].nodes_of_class)
            # print(dir(node.targets[0]))
            self.add_message("direct-status-assign", node=node)  # , confidence=HIGH)


def register(linter: PyLinter) -> None:
    """required method to auto register this checker"""

    linter.register_checker(DirectStatusAssignmentChecker(linter))



#! /usr/bin/env python
"""Consolidation of metadata based on OpenLibrary API as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.built_in.search_sources.open_library as open_library_connector
import colrev.ops.search_sources
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class OpenLibraryMetadataPrep(JsonSchemaMixin):
    """Prepares records based on openlibrary.org metadata"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "ask the publisher to correct the metadata"
    always_apply_changes = False

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)
        self.open_library_connector = open_library_connector.OpenLibrarySearchSource(
            source_operation=prep_operation
        )

    def check_availability(
        self, *, source_operation: colrev.operation.Operation
    ) -> None:
        """Check status (availability) of the Crossref API"""
        self.open_library_connector.check_availability(
            source_operation=source_operation
        )

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare the record metadata based on OpenLibrary"""

        if record.data.get("ENTRYTYPE", "NA") != "book":
            return record

        self.open_library_connector.get_masterdata(
            prep_operation=prep_operation, record=record
        )

        return record



#! /usr/bin/env python
"""CLI interface for manual preparation of PDFs"""
from __future__ import annotations

import os
import platform
import re
import webbrowser
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.record
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.ops.pdf_prep_man


# pylint: disable=too-few-public-methods


@zope.interface.implementer(
    colrev.env.package_manager.PDFPrepManPackageEndpointInterface
)
@dataclass
class CoLRevCLIPDFManPrep(JsonSchemaMixin):
    """Manually prepare PDFs based on a CLI (not yet implemented)"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = False

    __to_skip: int = 0

    def __init__(
        self,
        *,
        pdf_prep_man_operation: colrev.ops.pdf_prep_man.PDFPrepMan,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)
        self.review_manager = pdf_prep_man_operation.review_manager

    def __update_metadata(
        self, *, record: colrev.record.Record
    ) -> colrev.record.Record:
        valid_selections = ["a", "c", "t", "v", "n", "p", "s"]
        user_selection = ""
        print(
            "Update metadata fields: "
            "(a)uthor, (c)ontainer title, (t)itle, (v)olume, (n)umber, (p)ages  or (s)ave"
        )
        while user_selection not in valid_selections:
            user_selection = input("Selection: ")

            if user_selection == "s":
                break
            if user_selection == "a":
                author = input("Authors:")
                record.update_field(
                    key="author", value=author, source="manual_correction"
                )
            elif user_selection == "c":
                if "journal" in record.data:
                    journal = input("Journal:")
                    record.update_field(
                        key="journal", value=journal, source="manual_correction"
                    )
                if "booktitle" in record.data:
                    booktitle = input("Booktitle:")
                    record.update_field(
                        key="booktitle", value=booktitle, source="manual_correction"
                    )
            elif user_selection == "t":
                title = input("Title:")
                record.update_field(
                    key="title", value=title, source="manual_correction"
                )
            elif user_selection == "v":
                volume = input("Volume:")
                record.update_field(
                    key="volume", value=volume, source="manual_correction"
                )
            elif user_selection == "n":
                number = input("Number:")
                record.update_field(
                    key="number", value=number, source="manual_correction"
                )
            elif user_selection == "p":
                pages = input("Pages:")
                record.update_field(
                    key="pages", value=pages, source="manual_correction"
                )
            user_selection = ""

        return record

    def __open_pdf(self, *, filepath: Path) -> None:
        # pylint: disable=no-member
        webbrowser.open(str(filepath))

    def __remove_page(
        self,
        *,
        user_selection: str,
        filepath: Path,
        pdf_prep_man_operation: colrev.ops.pdf_prep_man.PDFPrepMan,
    ) -> None:
        if user_selection == "c":
            try:
                pdf_prep_man_operation.extract_coverpage(filepath=filepath)
            except colrev_exceptions.InvalidPDFException:
                pass
        elif user_selection == "l":
            try:
                pdf_prep_man_operation.extract_lastpage(filepath=filepath)
            except colrev_exceptions.InvalidPDFException:
                pass
        elif user_selection == "r":
            range_str = ""
            while not re.match(r"(\d)+-(\d)+", range_str):
                range_str = input('Page range to remove (e.g., "0-10"):')

            pages_to_exclude = list(
                range(
                    int(range_str[: range_str.find("-")]),
                    int(range_str[range_str.find("-") + 1 :]),
                )
            )
            try:
                pdf_prep_man_operation.extract_pages(
                    filepath=filepath, pages_to_remove=pages_to_exclude
                )
            except colrev_exceptions.InvalidPDFException:
                pass

    def __is_inside_wsl(self) -> bool:
        return "wsl" in platform.uname().release.lower()

    def __man_pdf_prep_item(
        self,
        *,
        filepath: Path,
        record: colrev.record.Record,
        pdf_prep_man_operation: colrev.ops.pdf_prep_man.PDFPrepMan,
    ) -> None:
        if not self.__is_inside_wsl():
            self.__open_pdf(filepath=filepath)

        # if PDF > 100 pages, we may check on which page we find the title & print
        intro_paragraph = (
            "Prepared?\n"
            "       (y)es, \n"
            "       (n)o/delete file,\n"
            "       (s)kip, (s10) to skip 10 records, or (q)uit,\n"
            "       (c)overpage remove, (l)ast page remove, (r)emove page range, "
            "(m)etadata needs to be updated\n"
        )
        print(intro_paragraph)
        user_selection = ""
        valid_selections = ["y", "n", "s", "q"]
        while user_selection not in valid_selections:
            user_selection = input("Selection: ")
            if user_selection.startswith("s"):
                if user_selection[1:].isdigit():
                    self.__to_skip = int(user_selection[1:])
                return
            if user_selection in ["c", "l", "r"]:
                self.__remove_page(
                    user_selection=user_selection,
                    filepath=filepath,
                    pdf_prep_man_operation=pdf_prep_man_operation,
                )
            elif user_selection == "y":
                pdf_prep_man_operation.set_pdf_man_prepared(record=record)
            elif user_selection == "n":
                record.remove_field(key="file")
                record.set_status(
                    target_state=colrev.record.RecordState.pdf_needs_manual_retrieval
                )
                if filepath.is_file():
                    filepath.unlink()
            elif user_selection == "m":
                self.__update_metadata(record=record)
                print(intro_paragraph)
            elif user_selection == "q":
                raise QuitPressedException()
            else:
                print("Invalid selection.")

    def __man_pdf_prep_item_init(
        self,
        *,
        pdf_prep_man_operation: colrev.ops.pdf_prep_man.PDFPrepMan,
        records: dict,
        item: dict,
        stat: str,
    ) -> dict:
        current_platform = platform.system()
        if current_platform in ["Linux", "Darwin"]:
            os.system("clear")
        else:
            os.system("cls")

        print(stat)
        record = colrev.record.Record(data=item)
        record.print_pdf_prep_man()

        record_dict = records[item["ID"]]
        record = colrev.record.Record(data=record_dict)
        if (
            colrev.record.RecordState.pdf_needs_manual_preparation
            != record_dict["colrev_status"]
        ):
            return record_dict

        file_provenance = record.get_field_provenance(key="file")
        print(
            "Manual preparation needed:"
            f" {colors.RED}{file_provenance['note']}{colors.END}"
        )

        filepath = self.review_manager.path / Path(record_dict["file"])
        if not filepath.is_file():
            filepath = self.review_manager.pdf_dir / f"{record_dict['ID']}.pdf"
        record.data.update(colrev_pdf_id=record.get_colrev_pdf_id(pdf_path=filepath))
        if filepath.is_file():
            self.__man_pdf_prep_item(
                filepath=filepath,
                record=record,
                pdf_prep_man_operation=pdf_prep_man_operation,
            )

        else:
            print(f'File does not exist ({record.data["ID"]})')

        self.review_manager.dataset.save_records_dict(records=records)

        return records

    def pdf_prep_man(
        self, pdf_prep_man_operation: colrev.ops.pdf_prep_man.PDFPrepMan, records: dict
    ) -> dict:
        """Prepare PDF manually based on a cli"""

        pdf_prep_man_operation.review_manager.logger.info(
            "Loading data for pdf_prep_man"
        )
        pdf_prep_man_data = pdf_prep_man_operation.get_data()
        records = pdf_prep_man_operation.review_manager.dataset.load_records_dict()

        for i, item in enumerate(pdf_prep_man_data["items"]):
            if self.__to_skip > 0:
                self.__to_skip -= 1
                continue
            try:
                stat = str(i + 1) + "/" + str(pdf_prep_man_data["nr_tasks"])
                records = self.__man_pdf_prep_item_init(
                    pdf_prep_man_operation=pdf_prep_man_operation,
                    records=records,
                    item=item,
                    stat=stat,
                )
            except QuitPressedException:
                break

        pdf_prep_man_operation.review_manager.dataset.save_records_dict(records=records)
        pdf_prep_man_operation.review_manager.dataset.add_record_changes()

        if pdf_prep_man_operation.pdfs_prepared_manually():
            if input("Create commit (y/n)?") == "y":
                pdf_prep_man_operation.review_manager.create_commit(
                    msg="Prepare PDFs manually",
                    manual_author=True,
                )
        else:
            pdf_prep_man_operation.review_manager.logger.info(
                "Prepare PDFs manually. Afterwards, use colrev pdf-get-man"
            )

        return records


class QuitPressedException(Exception):
    """Quit-pressed exception"""



#! /usr/bin/env python
"""Consolidation of metadata based on SemanticScholar API as a prep operation"""
from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TYPE_CHECKING

import requests
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.search_sources
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class SemanticScholarPrep(JsonSchemaMixin):
    """Prepares records based on SemanticScholar metadata"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = (
        "fill out the online form: "
        + "https://www.semanticscholar.org/faq#correct-error"
    )
    always_apply_changes = False

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)
        _, email = prep_operation.review_manager.get_committer()
        self.headers = {"user-agent": f"{__name__} (mailto:{email})"}
        self.session = prep_operation.review_manager.get_cached_session()

    def __get_record_from_item(
        self, *, item: dict, record_in: colrev.record.PrepRecord
    ) -> colrev.record.PrepRecord:
        # pylint: disable=too-many-branches
        retrieved_record: dict = {}
        if "authors" in item:
            authors_string = " and ".join(
                [author["name"] for author in item["authors"] if "name" in author]
            )
            authors_string = colrev.record.PrepRecord.format_author_field(
                input_string=authors_string
            )
            retrieved_record.update(author=authors_string)
        if "abstract" in item:
            retrieved_record.update(abstract=item["abstract"])
        if "doi" in item:
            if str(item["doi"]).lower() != "none":
                retrieved_record.update(doi=str(item["doi"]).upper())
        if "title" in item:
            retrieved_record.update(title=item["title"])
        if "year" in item:
            retrieved_record.update(year=item["year"])
        # Note: semantic scholar does not provide data on the type of venue.
        # we therefore use the original ENTRYTYPE
        if "venue" in item:
            if "journal" in record_in.data:
                retrieved_record.update(journal=item["venue"])
            if "booktitle" in record_in.data:
                retrieved_record.update(booktitle=item["venue"])
        if "url" in item:
            retrieved_record["colrev.semantic_scholar.id"] = item["url"]

        keys_to_drop = []
        for key, value in retrieved_record.items():
            retrieved_record[key] = str(value).replace("\n", " ").lstrip().rstrip()
            if value in ["", "None"] or value is None:
                keys_to_drop.append(key)
        for key in keys_to_drop:
            record_in.remove_field(key=key)

        record = colrev.record.PrepRecord(data=retrieved_record)
        return record

    def retrieve_record_from_semantic_scholar(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        url: str,
        record_in: colrev.record.PrepRecord,
    ) -> colrev.record.PrepRecord:
        """Prepare the record metadata based on SemanticScholar"""

        # prep_operation.review_manager.logger.debug(url)
        ret = self.session.request(
            "GET", url, headers=self.headers, timeout=prep_operation.timeout
        )
        ret.raise_for_status()

        data = json.loads(ret.text)
        items = data["data"]
        if len(items) == 0:
            return record_in
        if "paperId" not in items[0]:
            return record_in

        paper_id = items[0]["paperId"]
        record_retrieval_url = "https://api.semanticscholar.org/v1/paper/" + paper_id
        # prep_operation.review_manager.logger.debug(record_retrieval_url)
        ret_ent = self.session.request(
            "GET",
            record_retrieval_url,
            headers=self.headers,
            timeout=prep_operation.timeout,
        )
        ret_ent.raise_for_status()
        item = json.loads(ret_ent.text)

        record = self.__get_record_from_item(item=item, record_in=record_in)
        record.add_provenance_all(source=record_retrieval_url)

        return record

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare a record based on metadata from SemanticScholar"""

        same_record_type_required = (
            prep_operation.review_manager.settings.is_curated_masterdata_repo()
        )
        try:
            search_api_url = (
                "https://api.semanticscholar.org/graph/v1/paper/search?query="
            )
            url = search_api_url + record.data.get("title", "").replace(" ", "+")

            retrieved_record = self.retrieve_record_from_semantic_scholar(
                prep_operation=prep_operation, url=url, record_in=record
            )
            if "colrev.semantic_scholar.id" not in retrieved_record.data:
                return record

            # Remove fields that are not/rarely available before
            # calculating similarity metrics
            orig_record = record.copy_prep_rec()
            for key in ["volume", "number", "number", "pages"]:
                if key in orig_record.data:
                    record.remove_field(key=key)

            similarity = colrev.record.PrepRecord.get_retrieval_similarity(
                record_original=orig_record,
                retrieved_record_original=retrieved_record,
                same_record_type_required=same_record_type_required,
            )
            if similarity > prep_operation.retrieval_similarity:
                # prep_operation.review_manager.logger.debug("Found matching record")
                # prep_operation.review_manager.logger.debug(
                #     f"scholar similarity: {similarity} "
                #     f"(>{prep_operation.retrieval_similarity})"
                # )

                record.merge(
                    merging_record=retrieved_record,
                    default_source=retrieved_record.data["colrev.semantic_scholar.id"],
                )

            else:
                # prep_operation.review_manager.logger.debug(
                #     f"scholar similarity: {similarity} "
                #     f"(<{prep_operation.retrieval_similarity})"
                # )
                pass

        except (
            requests.exceptions.RequestException,
            colrev_exceptions.InvalidMerge,
        ):
            pass
        return record



#! /usr/bin/env python
"""SearchSource: SYNERGY-datasets"""
from __future__ import annotations

import datetime
import tempfile
import typing
from dataclasses import dataclass
from pathlib import Path

import pandas as pd
import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin
from git import Repo

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.load_utils_bib
import colrev.ops.search
import colrev.record
import colrev.ui_cli.cli_colors as colors

# pylint: disable=unused-argument

# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class SYNERGYDatasetsSearchSource(JsonSchemaMixin):
    """SearchSource for SYNERGY-datasets

    https://github.com/asreview/synergy-dataset

    csv files available in the releases (output directories) are supported:
    https://github.com/asreview/synergy-dataset/tags

    csv files containing article IDs are supported by source-specific preparation.
    metadata is added based on dois and pubmedid (openalex is not yet supported)

    """

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "ID"
    search_type = colrev.settings.SearchType.OTHER
    api_search_supported = True
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "SYNERGY-datasets"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/synergy_datasets.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        self.quality_model = source_operation.review_manager.get_qm()
        self.review_manager = source_operation.review_manager

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for SYNERGY-datasets"""

        result = {"confidence": 0.0}

        if "doi,pmid,openalex_id,label_included" in data:
            result["confidence"] = 1.0

        if '"doi","pmid","openalex_id","label_included"' in data:
            result["confidence"] = 1.0

        if (
            "record_id,pubmedID,title,authors,abstract,year,label_included,label_abstract_screening"
            in data
        ):
            result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        if params is None:
            operation.review_manager.logger.info("Retrieving available datasets")
            date_now_string = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            temp_path = tempfile.gettempdir() / Path(f"{date_now_string}-synergy")
            temp_path.mkdir()
            Repo.clone_from(
                "https://github.com/asreview/synergy-dataset", temp_path, depth=1
            )
            data_path = temp_path / Path("datasets")
            files = data_path.glob("**/*_ids.csv")
            print("https://github.com/asreview/synergy-dataset")
            print(
                "\n- "
                + "\n- ".join([str(f.parent.name) + "/" + str(f.name) for f in files])
            )
            params = input("Enter dataset:")
            params = "dataset=" + params

        if params.startswith("dataset="):
            dataset = params.replace("dataset=", "")
            filename = operation.get_unique_filename(
                file_path_string=f"SYNERGY_{dataset.replace('/', '_').replace('_ids.csv', '')}"
            )
            add_source = colrev.settings.SearchSource(
                endpoint="colrev.synergy_datasets",
                filename=filename,
                search_type=colrev.settings.SearchType.OTHER,
                search_parameters={"dataset": dataset},
                comment="",
            )
            operation.review_manager.settings.sources.append(add_source)
            return

        raise colrev_exceptions.PackageParameterError(
            f"Cannot add SYNERGY endpoint with query {params}"
        )

    def __load_dataset(self) -> pd.DataFrame:
        date_now_string = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        temp_path = tempfile.gettempdir() / Path(f"{date_now_string}-synergy")
        temp_path.mkdir()
        Repo.clone_from(
            "https://github.com/asreview/synergy-dataset", temp_path, depth=1
        )
        dataset_name = self.search_source.search_parameters["dataset"]
        dataset_df = pd.read_csv(temp_path / Path("datasets") / dataset_name)

        # check data structure
        if not all(x in dataset_df for x in ["doi", "pmid", "openalex_id"]):
            raise colrev_exceptions.SearchSourceException(
                f"Missing doi/pmid/openalex_id in {dataset_name}"
            )
        if "pmid" not in dataset_df:
            dataset_df["pmid"] = pd.NA
        missing_metadata = (
            dataset_df[["doi", "pmid", "openalex_id"]].isna().all(axis="columns")
        )
        missing_metadata_percentage = missing_metadata.sum() / dataset_df.shape[0]
        if missing_metadata_percentage > 0.1:
            self.review_manager.logger.error(
                f"Missing metadata percentage: {missing_metadata_percentage}"
            )
            input("ENTER to continue anyway")
        else:
            self.review_manager.logger.info(
                f"Missing metadata: {missing_metadata_percentage:.2%}"
            )
        return dataset_df

    def __validate_decisions(self, *, decisions: dict, record: dict) -> None:
        if "doi" in record:
            doi = record["doi"].lower()
            if doi in decisions["doi"]:
                decisions["doi"][doi].append(record["label_included"])
            else:
                decisions["doi"][doi] = [record["label_included"]]

        if "pmid" in record:
            pmid = record["pmid"].lower()
            if pmid in decisions["pmid"]:
                decisions["pmid"][pmid].append(record["label_included"])
            else:
                decisions["pmid"][pmid] = [record["label_included"]]

        if "openalex_id" in record:
            oaid = record["openalex_id"].lower()
            if oaid in decisions["openalex_id"]:
                decisions["openalex_id"][oaid].append(record["label_included"])
            else:
                decisions["openalex_id"][oaid] = [record["label_included"]]

    def __check_quality(self, *, decisions: dict) -> None:
        decisions["doi"] = {
            d: v for d, v in decisions["doi"].items() if len(v) > 1 and len(set(v)) != 1
        }
        decisions["pmid"] = {
            d: v
            for d, v in decisions["pmid"].items()
            if len(v) > 1 and len(set(v)) != 1
        }
        decisions["openalex_id"] = {
            d: v
            for d, v in decisions["openalex_id"].items()
            if len(v) > 1 and len(set(v)) != 1
        }
        if decisions["doi"] or decisions["pmid"] or decisions["openalex_id"]:
            self.review_manager.logger.error(
                "Errors in dataset: ambiguous inclusion decisions:"
            )
            msg = (
                f"{colors.RED}"
                + f"- dois: {', '.join(decisions['doi'])}"
                + f"- pmid: {', '.join(decisions['pmid'])}"
                + f"- openalex_id: {', '.join(decisions['openalex_id'])}"
                + f"{colors.END}"
            )
            if self.review_manager.force_mode:
                print(msg)
            else:
                raise colrev_exceptions.SearchSourceException(msg)

    def __prep_record(self, *, record: dict, ind: int) -> None:
        record["ID"] = ind
        record["ENTRYTYPE"] = "article"
        for k in list(record.keys()):
            if str(record[k]) == "nan":
                del record[k]
        if "doi" in record.keys():
            record["doi"] = str(record["doi"]).replace("https://doi.org/", "").upper()
        if "pmid" in record.keys():
            record["pmid"] = str(record["pmid"]).replace(
                "https://pubmed.ncbi.nlm.nih.gov/", ""
            )
        if "openalex_id" in record.keys():
            record["openalex_id"] = str(record["openalex_id"]).replace(
                "https://openalex.org/", ""
            )

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of the SYNERGY datasets"""

        dataset_df = self.__load_dataset()

        synergy_feed = self.search_source.get_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=False,
        )
        existing_keys = {
            "doi": [r["doi"] for r in synergy_feed.feed_records.values() if "doi" in r],
            "pmid": [
                r["pmid"] for r in synergy_feed.feed_records.values() if "pmid" in r
            ],
            "openalex_id": [
                r["openalex_id"]
                for r in synergy_feed.feed_records.values()
                if "openalex_id" in r
            ],
        }

        decisions: typing.Dict[str, typing.Dict[str, list]] = {
            "doi": {},
            "pmid": {},
            "openalex_id": {},
        }
        empty_records, duplicates = 0, 0
        for ind, record in enumerate(dataset_df.to_dict(orient="records")):
            self.__prep_record(record=record, ind=ind)
            self.__validate_decisions(decisions=decisions, record=record)
            # Skip records without metadata
            if {"ID", "ENTRYTYPE", "label_included"} == set(record.keys()):
                empty_records += 1
                continue

            # Skip records that are already in the feed
            if any(
                value in existing_keys[key]
                for key, value in record.items()
                if key in ["doi", "pmid", "openalex_id"]
            ):
                duplicates += 1
                continue

            for key in list(record.keys()):
                if key not in ["ID", "doi", "pmid", "openalex_id", "ENTRYTYPE"]:
                    record[f"colrev.synergy_datasets.{key}"] = record.pop(key)

            if "doi" in record:
                existing_keys["doi"].append(record["doi"])
            if "pmid" in record:
                existing_keys["pmid"].append(record["pmid"])
            if "openalex_id" in record:
                existing_keys["openalex_id"].append(record["openalex_id"])

            synergy_feed.set_id(record_dict=record)
            synergy_feed.add_record(record=colrev.record.Record(data=record))

            # The linking of doi/... should happen in the prep operation

        self.__check_quality(decisions=decisions)
        self.review_manager.logger.info(f"Dropped {empty_records} empty records")
        self.review_manager.logger.info(f"Dropped {duplicates} duplicate records")

        synergy_feed.save_feed_file()

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation,
                source=self.search_source,
                check_bib_file=False,
            )
            for record in records.values():
                if "pmid" in record:
                    record["pubmedid"] = record["pmid"].replace(
                        "https://pubmed.ncbi.nlm.nih.gov/", ""
                    )
                    del record["pmid"]
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for SYNERGY-datasets"""

        record.rename_field(
            key="colrev.synergy_datasets.pubmedid", new_key="colrev.pubmed.pubmedid"
        )
        record.rename_field(
            key="colrev.synergy_datasets.openalex_id", new_key="colrev.open_alex.id"
        )
        if not any(
            x in record.data
            for x in ["colrev.pubmed.pubmedid", "doi", "colrev.open_alex.id"]
        ):
            record.prescreen_exclude(reason="no-metadata-available")
        return record
